@@ -1,4 +1,6 @@
-/*
+/* -*- mode: c; c-basic-offset: 8; -*-
+ * vim: noexpandtab sw=8 ts=8 sts=0:
+ *
  * commands.c
  *
  * handles debugfs commands
@@ -70,6 +72,9 @@ static void do_fs_locks (char **args);
 static void do_bmap (char **args);
 static void do_icheck (char **args);
 static void do_dlm_locks (char **args);
+static void do_controld(char **args);
+static void do_dirblocks(char **args);
+static void do_xattr(char **args);
 
 dbgfs_gbls gbls;
 
@@ -79,6 +84,7 @@ static Command commands[] = {
 	{ "cd",		do_cd },
 	{ "chroot",	do_chroot },
 	{ "close",	do_close },
+	{ "controld",   do_controld },
 	{ "curdev",	do_curdev },
 	{ "dlm_locks",	do_dlm_locks },
 	{ "dump",	do_dump },
@@ -102,8 +108,10 @@ static Command commands[] = {
 	{ "slotmap",	do_slotmap },
 	{ "stat",	do_stat },
 	{ "stats",	do_stats },
+	{ "xattr",	do_xattr },
 	{ "encode",	do_encode_lockres },
-	{ "decode",	do_decode_lockres }
+	{ "decode",	do_decode_lockres },
+	{ "dirblocks",	do_dirblocks },
 };
 
 /*
@@ -824,14 +832,16 @@ static void do_help (char **args)
 	printf ("cd <filespec>\t\t\t\tChange directory\n");
 	printf ("chroot <filespec>\t\t\tChange root\n");
 	printf ("close\t\t\t\t\tClose a device\n");
+	printf ("controld dump\t\t\tObtain information from ocfs2_controld\n");
 	printf ("curdev\t\t\t\t\tShow current device\n");
 	printf ("decode <lockname#> ...\t\t\tDecode block#(s) from the lockname(s)\n");
-	printf ("dlm_locks [-l] lockname\t\t\tShow live dlm locking state\n");
+	printf ("dlm_locks [-f <file>] [-l] lockname\t\t\tShow live dlm locking state\n");
 	printf ("dump [-p] <filespec> <outfile>\t\tDumps file to outfile on a mounted fs\n");
+	printf ("dirblocks <filespec>\t\t\tDump directory blocks\n");
 	printf ("encode <filespec>\t\t\tShow lock name\n");
 	printf ("extent <block#>\t\t\t\tShow extent block\n");
 	printf ("findpath <block#>\t\t\tList one pathname of the inode/lockname\n");
-	printf ("fs_locks [-l] [-B]\t\t\tShow live fs locking state\n");
+	printf ("fs_locks [-f <file>] [-l] [-B]\t\t\tShow live fs locking state\n");
 	printf ("group <block#>\t\t\t\tShow chain group\n");
 	printf ("hb\t\t\t\t\tShows the used heartbeat blocks\n");
 	printf ("help, ?\t\t\t\t\tThis information\n");
@@ -847,6 +857,7 @@ static void do_help (char **args)
 	printf ("slotmap\t\t\t\t\tShow slot map\n");
 	printf ("stat <filespec>\t\t\t\tShow inode\n");
 	printf ("stats [-h]\t\t\t\tShow superblock\n");
+	printf ("xattr [-v] <filespec>\t\t\tShow Extended Attributes\n");
 }
 
 /*
@@ -890,6 +901,43 @@ static void do_lcd (char **args)
 	return ;
 }
 
+/*
+ * do_controld_dump()
+ *
+ */
+static void do_controld_dump(char **args)
+{
+	FILE *out;
+	errcode_t ret;
+	char *debug_buffer;
+
+	ret = o2cb_control_daemon_debug(&debug_buffer);
+	if (ret) {
+		com_err(args[0], ret, "while obtaining the debug buffer");
+		return;
+	}
+
+	out = open_pager(gbls.interactive);
+	fprintf(out, "%s", debug_buffer);
+	close_pager(out);
+	free(debug_buffer);
+}
+
+/*
+ * do_controld()
+ *
+ */
+static void do_controld(char **args)
+{
+	if (!args[1])
+		fprintf(stderr, "%s: Operation required\n", args[0]);
+	else if (!strcmp(args[1], "dump"))
+		do_controld_dump(args);
+	else
+		fprintf(stderr, "%s: Invalid operation: \"%s\"\n",
+			args[0], args[1]);
+}
+
 /*
  * do_curdev()
  *
@@ -998,7 +1046,7 @@ static void do_stat (char **args)
 		dump_fast_symlink(out, (char *)inode->id2.i_symlink);
 	else if (inode->i_flags & OCFS2_DEALLOC_FL)
 		dump_truncate_log(out, &(inode->id2.i_dealloc));
-	else
+	else if (!(inode->i_dyn_features & OCFS2_INLINE_DATA_FL))
 		ret = traverse_extents(gbls.fs, &(inode->id2.i_list), out);
 
 	if (ret)
@@ -1084,7 +1132,7 @@ static void do_dump (char **args)
 		return ;
 	}
 
-	fd = open(out_fn, O_CREAT | O_WRONLY | O_TRUNC, 0666);
+	fd = open64(out_fn, O_CREAT | O_WRONLY | O_TRUNC, 0666);
 	if (fd < 0) {
 		com_err(args[0], errno, "'%s'", out_fn);
 		return ;
@@ -1194,6 +1242,72 @@ static void do_group (char **args)
 	return ;
 }
 
+static int dirblocks_proxy(ocfs2_filesys *fs, uint64_t blkno,
+			   uint64_t bcount, uint16_t ext_flags,
+			   void *priv_data)
+{
+	errcode_t ret;
+	struct dirblocks_walk *ctxt = priv_data;
+
+	ret = ocfs2_read_dir_block(fs, ctxt->di, blkno, ctxt->buf);
+	if (!ret) {
+		fprintf(ctxt->out, "\tDirbloc: %"PRIu64"\n", blkno);
+		dump_dir_block(ctxt->out, ctxt->buf);
+	} else
+		com_err(gbls.cmd, ret,
+			"while reading dirblock %"PRIu64" on inode "
+			"%"PRIu64"\n",
+			blkno, ctxt->di->i_blkno);
+	return 0;
+}
+
+/*
+ * do_dirblocks()
+ *
+ */
+static void do_dirblocks (char **args)
+{
+	uint64_t ino_blkno;
+	errcode_t ret = 0;
+	struct dirblocks_walk ctxt = {
+		.fs = gbls.fs,
+	};
+
+	if (process_inode_args(args, &ino_blkno))
+		return;
+
+	ret = ocfs2_check_directory(gbls.fs, ino_blkno);
+	if (ret) {
+		com_err(args[0], ret, "while checking directory at "
+			"block %"PRIu64"", ino_blkno);
+		return;
+	}
+
+	ret = ocfs2_read_inode(gbls.fs, ino_blkno, gbls.blockbuf);
+	if (ret) {
+		com_err(args[0], ret, "while reading inode %"PRIu64"",
+			ino_blkno);
+		return;
+	}
+	ctxt.di = (struct ocfs2_dinode *)gbls.blockbuf;
+
+	ret = ocfs2_malloc_block(ctxt.fs->fs_io, &ctxt.buf);
+	if (ret) {
+		com_err(gbls.cmd, ret, "while allocating a block");
+		return;
+	}
+
+	ctxt.out = open_pager(gbls.interactive);
+	ret = ocfs2_block_iterate_inode(gbls.fs, ctxt.di, 0,
+					dirblocks_proxy, &ctxt);
+	if (ret)
+		com_err(args[0], ret, "while iterating directory at "
+			"block %"PRIu64"", ino_blkno);
+	close_pager(ctxt.out);
+
+	ocfs2_free(&ctxt.buf);
+}
+
 /*
  * do_extent()
  *
@@ -1357,52 +1471,39 @@ static void do_rdump(char **args)
 /*
  * do_encode_lockres()
  *
+ * This function only encodes the Super and the Inode lock. For the
+ * rest, use the --encode parameter directly.
  */
 static void do_encode_lockres (char **args)
 {
-	struct ocfs2_dinode *inode;
+	struct ocfs2_dinode *inode = (struct ocfs2_dinode *)gbls.blockbuf;
 	uint64_t blkno;
-	char *buf = NULL;
+	uint32_t gen = 0;
 	errcode_t ret = 0;
-	char suprlock[50] = "\0";
-	char metalock[50] = "\0";
-	char datalock[50] = "\0";
-	char rdwrlock[50] = "\0";
+	char lock[OCFS2_LOCK_ID_MAX_LEN];
+	enum ocfs2_lock_type type = OCFS2_LOCK_TYPE_META;
 
 	if (process_inode_args(args, &blkno))
-		return ;
+		return;
 
-	if (blkno == OCFS2_SUPER_BLOCK_BLKNO) {
-		ret = ocfs2_encode_lockres(OCFS2_LOCK_TYPE_SUPER, blkno, 0,
-					   suprlock);
-	} else {
-		buf = gbls.blockbuf;
-		ret = ocfs2_read_inode(gbls.fs, blkno, buf);
-		if (!ret) {
-			inode = (struct ocfs2_dinode *)buf;
-			ocfs2_encode_lockres(OCFS2_LOCK_TYPE_META, blkno,
-					     inode->i_generation, metalock);
-			ocfs2_encode_lockres(OCFS2_LOCK_TYPE_DATA, blkno,
-					     inode->i_generation, datalock);
-			ocfs2_encode_lockres(OCFS2_LOCK_TYPE_RW, blkno,
-					     inode->i_generation, rdwrlock);
+	if (blkno == OCFS2_SUPER_BLOCK_BLKNO)
+		type = OCFS2_LOCK_TYPE_SUPER;
+	else {
+		ret = ocfs2_read_inode(gbls.fs, blkno, (char *)inode);
+		if (ret) {
+			com_err(args[0], ret, "while reading inode %"PRIu64"",
+				blkno);
+			return;
 		}
+		gen = inode->i_generation;
 	}
 
-	if (ret) {
-		com_err(args[0], ret, "while reading inode %"PRIu64"", blkno);
-		return ;
-	}
+	ret = ocfs2_encode_lockres(type, blkno, gen, 0, lock);
+	if (ret)
+		return;
 
 	printf("\t");
-	if (*suprlock)
-		printf("%s ", suprlock);
-	if (*metalock)
-		printf("%s ", metalock);
-	if (*datalock)
-		printf("%s ", datalock);
-	if (*rdwrlock)
-		printf("%s ", rdwrlock);
+	printf("%s ", lock);
 	printf("\n");
 
 	return ;
@@ -1455,43 +1556,48 @@ static void do_dlm_locks(char **args)
 	FILE *out;
 	int dump_lvbs = 0;
 	int i;
-	struct locknames *lock;
 	struct list_head locklist;
-	struct list_head *iter, *iter2;
+	char *path = NULL, *uuid_str = NULL;
+	int c, argc;
 
-	if (check_device_open())
-		return;
+	init_stringlist(&locklist);
 
-	INIT_LIST_HEAD(&locklist);
+	for (argc = 0; (args[argc]); ++argc);
+	optind = 0;
 
-	i = 1;
-	if (args[i] && strlen(args[i])) {
-		if (!strcmp("-l", args[i])) {
+	while ((c = getopt(argc, args, "lf:")) != -1) {
+		switch (c) {
+		case 'l':
 			dump_lvbs = 1;
-			i++;
+			break;
+		case 'f':
+			path = optarg;
+			break;
+		default:
+			break;
 		}
+	}
+	if ((path == NULL)) {
+		/* Only error for a missing device if we're asked to
+		 * read from a live file system. */
+		if (check_device_open())
+			return;
 
-		for ( ; args[i] && strlen(args[i]); ++i) {
-			lock = calloc(1, sizeof(struct locknames));
-			if (!lock)
+		uuid_str = gbls.fs->uuid_str;
+	}
+
+	i = optind;
+	if (args[i] && strlen(args[i])) {
+		for ( ; args[i] && strlen(args[i]); ++i)
+			if (add_to_stringlist(args[i], &locklist))
 				break;
-			INIT_LIST_HEAD(&lock->list);
-			strncpy(lock->name, args[i], sizeof(lock->name));
-			list_add_tail(&lock->list, &locklist);
-		}
 	}
 
 	out = open_pager(gbls.interactive);
-	dump_dlm_locks(gbls.fs->uuid_str, out, dump_lvbs, &locklist);
+	dump_dlm_locks(uuid_str, out, path, dump_lvbs, &locklist);
 	close_pager(out);
 
-	if (!list_empty(&locklist)) {
-		list_for_each_safe(iter, iter2, &locklist) {
-			lock = list_entry(iter, struct locknames, list);
-			list_del(iter);
-			free(lock);
-		}
-	}
+	free_stringlist(&locklist);
 }
 
 /*
@@ -1503,21 +1609,51 @@ static void do_fs_locks(char **args)
 	FILE *out;
 	int dump_lvbs = 0;
 	int only_busy = 0;
-	int i = 0;
+	int c, argc;
+	struct list_head locklist;
+	char *path = NULL, *uuid_str = NULL;
 
-	if (check_device_open())
-		return;
+	for (argc = 0; (args[argc]); ++argc);
+	optind = 0;
 
-	while (args[++i]) {
-		if (!strcmp("-l", args[i]))
+	while ((c = getopt(argc, args, "lBf:")) != -1) {
+		switch (c) {
+		case 'l':
 			dump_lvbs = 1;
-		else if (!strcmp("-B", args[i]))
+			break;
+		case 'B':
 			only_busy = 1;
+			break;
+		case 'f':
+			path = optarg;
+			break;
+		default:
+			break;
+		}
+	}
+	if ((path == NULL)) {
+		/* Only error for a missing device if we're asked to
+		 * read from a live file system. */
+		if (check_device_open())
+			return;
+
+		uuid_str = gbls.fs->uuid_str;
+	}
+
+	init_stringlist(&locklist);
+
+	if (optind < argc) {
+		for ( ; args[optind] && strlen(args[optind]); ++optind)
+			if (add_to_stringlist(args[optind], &locklist))
+				break;
 	}
 
 	out = open_pager(gbls.interactive);
-	dump_fs_locks(gbls.fs->uuid_str, out, dump_lvbs, only_busy);
+	dump_fs_locks(uuid_str, out, path, dump_lvbs, only_busy,
+		      &locklist);
 	close_pager(out);
+
+	free_stringlist(&locklist);
 }
 
 /*
@@ -1636,3 +1772,72 @@ static void do_icheck(char **args)
 
 	return;
 }
+/*
+ * do_xattr()
+ *
+ */
+static void do_xattr(char **args)
+{
+	struct ocfs2_dinode *inode;
+	char *buf = NULL;
+	FILE *out;
+	char *usage = "usage: xattr [-v] <filespec>";
+	uint64_t blkno, xattrs_bucket = 0;
+	uint32_t xattrs_ibody = 0, xattrs_block = 0;
+	int ind = 1, verbose = 0;
+	errcode_t ret = 0;
+
+	if (check_device_open())
+		return;
+
+	if (!args[1]) {
+		fprintf(stderr, "%s\n", usage);
+		return;
+	}
+
+	if (!strcmp(args[1], "-v")) {
+		++ind;
+		++verbose;
+	}
+
+	if (!args[ind]) {
+		fprintf(stderr, "%s\n", usage);
+		return;
+	}
+
+	ret = string_to_inode(gbls.fs, gbls.root_blkno, gbls.cwd_blkno,
+			      args[ind], &blkno);
+	if (ret) {
+		com_err(args[0], ret, "while translating %s", args[ind]);
+		return;
+	}
+
+	buf = gbls.blockbuf;
+	ret = ocfs2_read_inode(gbls.fs, blkno, buf);
+	if (ret) {
+		com_err(args[0], ret, "while reading inode %"PRIu64"", blkno);
+		return;
+	}
+
+	inode = (struct ocfs2_dinode *)buf;
+	if (!inode->i_dyn_features & OCFS2_HAS_XATTR_FL)
+		return;
+
+	out = open_pager(gbls.interactive);
+
+	xattrs_ibody = dump_xattr_ibody(out, gbls.fs, inode, verbose);
+
+	if (inode->i_xattr_loc)
+		ret = dump_xattr_block(out, gbls.fs, inode, &xattrs_block,
+				       &xattrs_bucket, verbose);
+
+	if (ret)
+		com_err(args[0], ret, "while traversing inode at block "
+			"%"PRIu64, blkno);
+	else
+		fprintf(out, "\n\tExtended attributes in total: %"PRIu64"\n",
+			xattrs_ibody + xattrs_block + xattrs_bucket);
+	close_pager(out);
+
+	return ;
+}
