@@ -1,4 +1,6 @@
-/*
+/* -*- mode: c; c-basic-offset: 8; -*-
+ * vim: noexpandtab sw=8 ts=8 sts=0:
+ *
  * dump.c
  *
  * dumps ocfs2 structures
@@ -88,11 +90,16 @@ void dump_super_block(FILE *out, struct ocfs2_super_block *sb)
 
 	fprintf(out, "\tMax Node Slots: %u\n", sb->s_max_slots);
 
-	fprintf(out, "\tLabel: %s\n", sb->s_label);
+	fprintf(out, "\tExtended Attributes Inline Size: %u\n",
+		sb->s_xattr_inline_size);
+
+	fprintf(out, "\tLabel: %.*s\n", OCFS2_MAX_VOL_LABEL_LEN, sb->s_label);
 	fprintf(out, "\tUUID: ");
 	for (i = 0; i < 16; i++)
 		fprintf(out, "%02X", sb->s_uuid[i]);
 	fprintf(out, "\n");
+	fprintf(out, "\tUUID_hash: %u (0x%x)\n", sb->s_uuid_hash,
+		sb->s_uuid_hash);
 	if (ocfs2_userspace_stack(sb))
 		fprintf(out,
 			"\tCluster stack: %s\n"
@@ -154,6 +161,16 @@ void dump_fast_symlink (FILE *out, char *link)
 	return ;
 }
 
+/*
+ * dump_block_check
+ *
+ */
+void dump_block_check(FILE *out, struct ocfs2_block_check *bc)
+{
+	fprintf(out, "\tCRC32: %.8"PRIx32"   ECC: %.4"PRIx16"\n",
+		le32_to_cpu(bc->bc_crc32e), le16_to_cpu(bc->bc_ecc));
+}
+
 /*
  * dump_inode()
  *
@@ -165,6 +182,7 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 	char *str;
 	uint16_t mode;
 	GString *flags = NULL;
+	GString *dyn_features = NULL;
 	char tmp_str[30];
 	time_t tm;
 
@@ -213,15 +231,35 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 	if (in->i_flags & OCFS2_DEALLOC_FL)
 		g_string_append (flags, "Dealloc ");
 
+	dyn_features = g_string_new(NULL);
+	if (in->i_dyn_features & OCFS2_INLINE_DATA_FL)
+		g_string_append(dyn_features, "InlineData ");
+	if (in->i_dyn_features & OCFS2_HAS_XATTR_FL)
+		g_string_append(dyn_features, "HasXattr ");
+	if (in->i_dyn_features & OCFS2_INLINE_XATTR_FL)
+		g_string_append(dyn_features, "InlineXattr ");
+	if (in->i_dyn_features & OCFS2_INDEXED_DIR_FL)
+		g_string_append(dyn_features, "IndexedDir ");
+
 	fprintf(out, "\tInode: %"PRIu64"   Mode: 0%0o   Generation: %u (0x%x)\n",
 		(uint64_t)in->i_blkno, mode, in->i_generation, in->i_generation);
 
 	fprintf(out, "\tFS Generation: %u (0x%x)\n", in->i_fs_generation,
 		in->i_fs_generation);
 
+	dump_block_check(out, &in->i_check);
+
 	fprintf(out, "\tType: %s   Attr: 0x%x   Flags: %s\n", str, in->i_attr,
 		flags->str);
 
+	fprintf(out, "\tDynamic Features: (0x%x) %s\n", in->i_dyn_features,
+		dyn_features->str);
+
+	if (in->i_dyn_features & OCFS2_HAS_XATTR_FL)
+		fprintf(out, "\tExtended Attributes Block: %"PRIu64
+			"  Extended Attributes Inline Size: %u\n",
+			(uint64_t)in->i_xattr_loc, in->i_xattr_inline_size);
+
 	pw = getpwuid(in->i_uid);
 	gr = getgrgid(in->i_gid);
 	fprintf(out, "\tUser: %d (%s)   Group: %d (%s)   Size: %"PRIu64"\n",
@@ -269,8 +307,15 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 			in->id1.journal1.ij_recovery_generation);
 	}
 
+	if (in->i_dyn_features & OCFS2_INLINE_DATA_FL) {
+		fprintf(out, "\tInline Data Max: %u\n",
+			in->id2.i_data.id_count);
+	}
+
 	if (flags)
 		g_string_free (flags, 1);
+	if (dyn_features)
+		g_string_free(dyn_features, 1);
 	return ;
 }
 
@@ -357,6 +402,8 @@ void dump_extent_block (FILE *out, struct ocfs2_extent_block *blk)
 	fprintf (out, "\tBlknum: %"PRIu64"   Next Leaf: %"PRIu64"\n",
 		 (uint64_t)blk->h_blkno, (uint64_t)blk->h_next_leaf_blk);
 
+	dump_block_check(out, &blk->h_check);
+
 	return ;
 }
 
@@ -375,6 +422,8 @@ void dump_group_descriptor (FILE *out, struct ocfs2_group_desc *grp,
 			 grp->bg_chain,
 			 (uint64_t)grp->bg_parent_dinode,
 			 grp->bg_generation);
+		dump_block_check(out, &grp->bg_check);
+
 		fprintf(out, "\t##   %-15s   %-6s   %-6s   %-6s   %-6s   %-6s\n",
 			"Block#", "Total", "Used", "Free", "Contig", "Size");
 	}
@@ -427,6 +476,50 @@ int  dump_dir_entry (struct ocfs2_dir_entry *rec, int offset, int blocksize,
 	return 0;
 }
 
+/*
+ * dump_dir_block()
+ *
+ */
+void dump_dir_block(FILE *out, char *buf)
+{
+
+	struct ocfs2_dir_entry *dirent;
+	int offset = 0;
+	int end = ocfs2_dir_trailer_blk_off(gbls.fs);
+	struct ocfs2_dir_block_trailer *trailer =
+		ocfs2_dir_trailer_from_block(gbls.fs, buf);
+	list_dir_opts ls_opts = {
+		.fs	= gbls.fs,
+		.out	= out,
+	};
+
+	if (!strncmp((char *)trailer->db_signature, OCFS2_DIR_TRAILER_SIGNATURE,
+		     sizeof(trailer->db_signature))) {
+		fprintf(out,
+			"\tTrailer Block: %-15"PRIu64" Inode: %-15"PRIu64" rec_len: %-4u\n",
+			trailer->db_blkno, trailer->db_parent_dinode,
+			trailer->db_compat_rec_len);
+		dump_block_check(out, &trailer->db_check);
+	} else
+		end = gbls.fs->fs_blocksize;
+
+	fprintf(out, "\tEntries:\n");
+	while (offset < end) {
+		dirent = (struct ocfs2_dir_entry *) (buf + offset);
+		if (((offset + dirent->rec_len) > end) ||
+		    (dirent->rec_len < 8) ||
+		    ((dirent->rec_len % 4) != 0) ||
+		    (((dirent->name_len & 0xFF)+8) > dirent->rec_len)) {
+			/* Corrupted */
+			return;
+		}
+
+		dump_dir_entry(dirent, offset, gbls.fs->fs_blocksize, NULL,
+			       &ls_opts);
+		offset += dirent->rec_len;
+	}
+}
+
 /*
  * dump_jbd_header()
  *
@@ -492,7 +585,8 @@ void dump_jbd_superblock (FILE *out, journal_superblock_t *jsb)
  * dump_jbd_block()
  *
  */
-void dump_jbd_block (FILE *out, journal_header_t *header, uint64_t blknum)
+void dump_jbd_block (FILE *out, journal_superblock_t *jsb,
+		     journal_header_t *header, uint64_t blknum)
 {
 	int i;
 	int j;
@@ -505,33 +599,36 @@ void dump_jbd_block (FILE *out, journal_header_t *header, uint64_t blknum)
 	uint32_t *blocknr;
 	char *uuid;
 	struct ocfs2_super_block *sb = OCFS2_RAW_SB(gbls.fs->fs_super);
+	int tag_bytes = ocfs2_journal_tag_bytes(jsb);
 
 	tagflg = g_string_new (NULL);
 
 	fprintf (out, "\tBlock %"PRIu64": ", blknum);
 
 	switch (ntohl(header->h_blocktype)) {
-	case JFS_DESCRIPTOR_BLOCK:
+	case JBD2_DESCRIPTOR_BLOCK:
 		fprintf (out, "Journal Descriptor\n");
 		dump_jbd_header (out, header);
 
 		fprintf (out, "\t%3s %-15s %-s\n", "No.", "Blocknum", "Flags");
 
 		for (i = sizeof(journal_header_t); i < (1 << sb->s_blocksize_bits);
-		     i+=sizeof(journal_block_tag_t)) {
+		     i+=tag_bytes) {
 			tag = (journal_block_tag_t *) &blk[i];
 
 			get_tag_flag(ntohl(tag->t_flags), tagflg);
-			fprintf (out, "\t%2d. %-15u %-s\n",
-				 count, ntohl(tag->t_blocknr), tagflg->str);
+			fprintf (out, "\t%2d. %-15"PRIu64" %-s\n",
+				 count,
+                                 ocfs2_journal_tag_block(tag, tag_bytes),
+                                 tagflg->str);
 			g_string_truncate (tagflg, 0);
 
-			if (tag->t_flags & htonl(JFS_FLAG_LAST_TAG))
+			if (tag->t_flags & htonl(JBD2_FLAG_LAST_TAG))
 				break;
 
 			/* skip the uuid. */
-			if (!(tag->t_flags & htonl(JFS_FLAG_SAME_UUID))) {
-				uuid = &blk[i + sizeof(journal_block_tag_t)];
+			if (!(tag->t_flags & htonl(JBD2_FLAG_SAME_UUID))) {
+				uuid = &blk[i + tag_bytes];
 				fprintf (out, "\tUUID: ");
 				for(j = 0; j < 16; j++)
 					fprintf (out, "%02X",uuid[j]);
@@ -542,12 +639,12 @@ void dump_jbd_block (FILE *out, journal_header_t *header, uint64_t blknum)
 		}
 		break;
 
-	case JFS_COMMIT_BLOCK:
+	case JBD2_COMMIT_BLOCK:
 		fprintf(out, "Journal Commit Block\n");
 		dump_jbd_header (out, header);
 		break;
 
-	case JFS_REVOKE_BLOCK:							/*TODO*/
+	case JBD2_REVOKE_BLOCK:							/*TODO*/
 		fprintf(out, "Journal Revoke Block\n");
 		dump_jbd_header (out, header);
 		revoke = (journal_revoke_header_t *) blk;
@@ -576,25 +673,45 @@ void dump_jbd_block (FILE *out, journal_header_t *header, uint64_t blknum)
  * dump_jbd_metadata()
  *
  */
-void dump_jbd_metadata (FILE *out, int type, char *buf, uint64_t blknum)
+void dump_jbd_metadata (FILE *out, enum dump_block_type type, char *buf,
+			uint64_t blknum)
 {
+	struct ocfs2_dir_block_trailer *trailer;
+
 	fprintf (out, "\tBlock %"PRIu64": ", blknum);
 	switch (type) {
-	case 1:
+	case DUMP_BLOCK_INODE:
 		fprintf(out, "Inode\n");
+		ocfs2_swap_inode_to_cpu((struct ocfs2_dinode *)buf,
+					gbls.fs->fs_blocksize);
 		dump_inode (out, (struct ocfs2_dinode *)buf);
 		fprintf (out, "\n");
 		break;
-	case 2:
+	case DUMP_BLOCK_EXTENT_BLOCK:
 		fprintf(out, "Extent\n");
+		ocfs2_swap_extent_block_to_cpu((struct ocfs2_extent_block *)buf);
 		dump_extent_block (out, (struct ocfs2_extent_block *)buf);
 		fprintf (out, "\n");
 		break;
-	case 3:
+	case DUMP_BLOCK_GROUP_DESCRIPTOR:
 		fprintf(out, "Group\n");
+		ocfs2_swap_group_desc((struct ocfs2_group_desc *)buf);
 		dump_group_descriptor (out, (struct ocfs2_group_desc *)buf, 0);
 		fprintf (out, "\n");
 		break;
+	case DUMP_BLOCK_DIR_BLOCK:
+		fprintf(out, "Dirblock\n");
+		/*
+		 * We know there's a trailer, because that's how it
+		 * was detected
+		 */
+		ocfs2_swap_dir_entries_to_cpu(buf,
+					      ocfs2_dir_trailer_blk_off(gbls.fs));
+		trailer = ocfs2_dir_trailer_from_block(gbls.fs, buf);
+		ocfs2_swap_dir_trailer(trailer);
+		dump_dir_block(out, buf);
+		fprintf (out, "\n");
+		break;
 	default:
 		fprintf (out, "TODO\n\n");
 		break;
@@ -711,3 +828,186 @@ void dump_icheck(FILE *out, int hdr, uint64_t blkno, uint64_t inode,
 
 	fprintf(out, "\t%-15"PRIu64"   %-15s   %-15s\n", blkno, inostr, offstr);
 }
+
+static void dump_xattr(FILE *out, struct ocfs2_xattr_header *xh)
+{
+	int i;
+
+	fprintf(out, "\t###   %-4s   %-6s  %-13s  %s\n", "Type",
+		"Inline", "Name Length", "Value Length");
+	for (i = 0 ; i < xh->xh_count; i++) {
+		struct ocfs2_xattr_entry *xe = &xh->xh_entries[i];
+
+		fprintf(out,
+			"\t#%-2d   %-4u   %-6u  %-13u  %-13"PRIu64"\n",
+			i, ocfs2_xattr_get_type(xe),
+			ocfs2_xattr_is_local(xe) ? 1 : 0,
+			xe->xe_name_len, (uint64_t)xe->xe_value_size);
+
+		if (!ocfs2_xattr_is_local(xe)) {
+			struct ocfs2_xattr_value_root *xv =
+				(struct ocfs2_xattr_value_root *)
+				((void *)xh + xe->xe_name_offset +
+				OCFS2_XATTR_SIZE(xe->xe_name_len));
+			struct ocfs2_extent_list *el = &xv->xr_list;
+			dump_extent_list(out, el);
+		}
+	}
+	return;
+}
+
+static errcode_t dump_xattr_buckets(FILE *out,
+				    ocfs2_filesys *fs,
+				    uint64_t blkno,
+				    uint32_t clusters,
+				    uint64_t *xattrs_bucket,
+				    int verbose)
+{
+	int i;
+	errcode_t ret = 0;
+	char *bucket = NULL;
+	struct ocfs2_xattr_header *xh;
+	int blk_per_bucket = ocfs2_blocks_per_xattr_bucket(fs);
+	uint32_t bpc = ocfs2_xattr_buckets_per_cluster(fs);
+	uint32_t num_buckets = clusters * bpc;
+
+	ret = ocfs2_malloc_blocks(fs->fs_io, blk_per_bucket, &bucket);
+	if (ret)
+		goto out;
+
+	fprintf(out, "\tExtended Attributes extent record start at #%"PRIu64
+		"  Has clusters: %u", blkno, clusters);
+	for (i = 0; i < num_buckets; i++, blkno += blk_per_bucket) {
+		ret = ocfs2_read_xattr_bucket(fs, blkno, bucket);
+		if (ret)
+			goto out;
+
+		xh = (struct ocfs2_xattr_header *)bucket;
+		/*
+		 * The real bucket num in this series of blocks is stored
+		 * in the 1st bucket.
+		 */
+		if (i == 0) {
+			num_buckets = xh->xh_num_buckets;
+			fprintf(out, "  Has buckets: %d\n", num_buckets);
+		}
+		fprintf(out, "\t\tExtended Attributes in bucket #%d: %u\n",
+			i, xh->xh_count);
+		if (verbose)
+			dump_xattr(out, xh);
+		*xattrs_bucket += xh->xh_count;
+	}
+
+out:
+	if (bucket)
+		ocfs2_free(&bucket);
+
+	return ret;
+}
+
+
+static errcode_t dump_xattr_index_block(FILE *out,
+					ocfs2_filesys *fs,
+					struct ocfs2_dinode *di,
+					struct ocfs2_xattr_block *xb,
+					uint64_t *xattrs_bucket,
+					int verbose)
+{
+	struct ocfs2_extent_list *el = &xb->xb_attrs.xb_root.xt_list;
+	errcode_t ret = 0;
+	uint32_t name_hash = UINT_MAX, e_cpos = 0, num_clusters = 0;
+	uint64_t p_blkno = 0;
+
+	if (!el->l_next_free_rec)
+		return 0;
+
+	fprintf(out, "\tExtended Attributes extent tree in index block #%"
+		PRIu64"  Depth: %d  Records: %d\n", (uint64_t)di->i_xattr_loc,
+		el->l_tree_depth, el->l_next_free_rec);
+	if (verbose)
+		dump_extent_list(out, el);
+
+	while (name_hash > 0) {
+		ret = ocfs2_xattr_get_rec(fs, xb, name_hash, &p_blkno,
+					  &e_cpos, &num_clusters);
+		if (ret)
+			goto out;
+
+		ret = dump_xattr_buckets(out, fs, p_blkno, num_clusters,
+					 xattrs_bucket, verbose);
+		if (ret)
+			goto out;
+
+		if (e_cpos == 0)
+			break;
+
+		name_hash = e_cpos - 1;
+	}
+
+out:
+	return ret;
+}
+
+/*
+ * dump_xattr_block()
+ *
+ */
+errcode_t dump_xattr_block(FILE *out, ocfs2_filesys *fs,
+			   struct ocfs2_dinode *in,
+			   uint32_t *xattrs_block,
+			   uint64_t *xattrs_bucket,
+			   int verbose)
+{
+	errcode_t ret;
+	char *blk = NULL;
+	struct ocfs2_xattr_block *xb = NULL;
+
+	ret = ocfs2_malloc_block(fs->fs_io, &blk);
+	if (ret)
+		goto out;
+
+	ret = ocfs2_read_xattr_block(fs, in->i_xattr_loc, blk);
+	if (ret)
+		goto out;
+
+	xb = (struct ocfs2_xattr_block *)blk;
+
+	if (!(xb->xb_flags & OCFS2_XATTR_INDEXED)) {
+		struct ocfs2_xattr_header *xh = &xb->xb_attrs.xb_header;
+		*xattrs_block = xh->xh_count;
+		fprintf(out, "\tExtended Attributes in block #%"PRIu64": %u\n",
+			(uint64_t)in->i_xattr_loc, *xattrs_block);
+		if (verbose)
+			dump_xattr(out, xh);
+	} else {
+		ret = dump_xattr_index_block(out, fs, in, xb,
+					     xattrs_bucket, verbose);
+	}
+
+out:
+	if (blk)
+		ocfs2_free(&blk);
+	return ret;
+}
+/*
+ * dump_xattr_ibody()
+ *
+ */
+uint32_t dump_xattr_ibody(FILE *out, ocfs2_filesys *fs,
+			  struct ocfs2_dinode *in, int verbose)
+{
+	if (in->i_dyn_features & OCFS2_INLINE_XATTR_FL) {
+		struct ocfs2_xattr_header *xh = (struct ocfs2_xattr_header *)
+						((void *)in + fs->fs_blocksize
+						- in->i_xattr_inline_size);
+
+		fprintf(out, "\tExtended Attributes inline: %u\n",
+			xh->xh_count);
+		if (verbose)
+			dump_xattr(out, xh);
+		return xh->xh_count;
+	} else {
+		fprintf(out, "\tExtended Attributes inline: 0\n");
+		return 0;
+	}
+}
