@@ -1,4 +1,6 @@
-/*
+/* -*- mode: c; c-basic-offset: 8; -*-
+ * vim: noexpandtab sw=8 ts=8 sts=0:
+ *
  * utils.c
  *
  * utility functions
@@ -27,85 +29,65 @@
 
 extern dbgfs_gbls gbls;
 
-/*
- * Tests features in a flag field and adds to a printable string.  Expects
- * the flag field named 'flag' and the string named 'str'.  Primarily for
- * readability.
- */
-#define test_feature(_flag, _flagstr) do {		\
-	typeof(flag) _f = (_flag);			\
-	if (flag & _f) {				\
-		g_string_append(str, (_flagstr));	\
-		g_string_append_c(str, ' ');		\
-		flag &= ~_f;				\
-	}						\
-} while (0)
-
 void get_incompat_flag(uint32_t flag, GString *str)
 {
+	errcode_t err;
+	char buf[PATH_MAX];
+	ocfs2_fs_options flags = {
+		.opt_incompat = flag,
+	};
 
-	test_feature(OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV, "Heartbeat");
-	test_feature(OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG, "AbortedResize");
-	test_feature(OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT, "Local");
-	test_feature(OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC, "Sparse");
-	test_feature(OCFS2_FEATURE_INCOMPAT_EXTENDED_SLOT_MAP,
-		     "ExtendedSlotMap");
-	test_feature(OCFS2_FEATURE_INCOMPAT_USERSPACE_STACK,
-		     "UserspaceClusterStack");
-	test_feature(OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG, "AbortedTunefs");
-
-        /* test_feature() clears out known flags */
-	if (flag)
-		g_string_append(str, "Unknown ");
-
-	if (!str->len)
-		g_string_append(str, "None");
-
-	return;
+	*buf = '\0';
+	err = ocfs2_snprint_feature_flags(buf, PATH_MAX, &flags);
+	if (!err)
+		g_string_append(str, buf);
+	else
+		com_err(gbls.cmd, err, "while processing feature flags");
 }
 
 void get_tunefs_flag(uint32_t incompat_flag, uint16_t flag, GString *str)
 {
-	if (!(incompat_flag & OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG)) {
-		g_string_append(str, "None");
-		return;
-	}
+	errcode_t err;
+	char buf[PATH_MAX];
 
-	if (flag & OCFS2_TUNEFS_INPROG_REMOVE_SLOT)
-		g_string_append(str, "RemoveSlot ");
-
-	if (flag & ~OCFS2_TUNEFS_INPROG_REMOVE_SLOT)
-		g_string_append(str, "Unknown ");
-
-	return;
+	*buf = '\0';
+	err = ocfs2_snprint_tunefs_flags(buf, PATH_MAX, flag);
+	if (!err)
+		g_string_append(str, buf);
+	else
+		com_err(gbls.cmd, err, "while processing inprog flags");
 }
 
 void get_compat_flag(uint32_t flag, GString *str)
 {
-	test_feature(OCFS2_FEATURE_COMPAT_BACKUP_SB, "BackupSuper");
+	errcode_t err;
+	char buf[PATH_MAX];
+	ocfs2_fs_options flags = {
+		.opt_compat = flag,
+	};
 
-        /* test_feature() clears out known flags */
-        if (flag)
-		g_string_append(str, "Unknown ");
-
-	if (!str->len)
-		g_string_append(str, "None");
-
-	return;
+	*buf = '\0';
+	err = ocfs2_snprint_feature_flags(buf, PATH_MAX, &flags);
+	if (!err)
+		g_string_append(str, buf);
+	else
+		com_err(gbls.cmd, err, "while processing feature flags");
 }
 
 void get_rocompat_flag(uint32_t flag, GString *str)
 {
-	test_feature(OCFS2_FEATURE_RO_COMPAT_UNWRITTEN, "Unwritten");
+	errcode_t err;
+	char buf[PATH_MAX];
+	ocfs2_fs_options flags = {
+		.opt_ro_compat = flag,
+	};
 
-        /* test_feature() clears out known flags */
-        if (flag)
-		 g_string_append(str, "Unknown ");
-
-	if (!str->len)
-		g_string_append(str, "None");
-
-	return;
+	*buf = '\0';
+	err = ocfs2_snprint_feature_flags(buf, PATH_MAX, &flags);
+	if (!err)
+		g_string_append(str, buf);
+	else
+		com_err(gbls.cmd, err, "while processing feature flags");
 }
 
 /*
@@ -233,20 +215,20 @@ void get_publish_flag (uint32_t flag, GString *str)
 void get_journal_block_type (uint32_t jtype, GString *str)
 {
 	switch (jtype) {
-	case JFS_DESCRIPTOR_BLOCK:
-		g_string_append (str, "JFS_DESCRIPTOR_BLOCK");
+	case JBD2_DESCRIPTOR_BLOCK:
+		g_string_append (str, "JBD2_DESCRIPTOR_BLOCK");
 		break;
-	case JFS_COMMIT_BLOCK:
-		g_string_append (str, "JFS_COMMIT_BLOCK");
+	case JBD2_COMMIT_BLOCK:
+		g_string_append (str, "JBD2_COMMIT_BLOCK");
 		break;
-	case JFS_SUPERBLOCK_V1:
-		g_string_append (str, "JFS_SUPERBLOCK_V1");
+	case JBD2_SUPERBLOCK_V1:
+		g_string_append (str, "JBD2_SUPERBLOCK_V1");
 		break;
-	case JFS_SUPERBLOCK_V2:
-		g_string_append (str, "JFS_SUPERBLOCK_V2");
+	case JBD2_SUPERBLOCK_V2:
+		g_string_append (str, "JBD2_SUPERBLOCK_V2");
 		break;
-	case JFS_REVOKE_BLOCK:
-		g_string_append (str, "JFS_REVOKE_BLOCK");
+	case JBD2_REVOKE_BLOCK:
+		g_string_append (str, "JBD2_REVOKE_BLOCK");
 		break;
 	}
 
@@ -267,17 +249,17 @@ void get_tag_flag (uint32_t flags, GString *str)
 		goto done;
 	}
 
-	if (flags & JFS_FLAG_ESCAPE)
-		g_string_append (str, "JFS_FLAG_ESCAPE ");
+	if (flags & JBD2_FLAG_ESCAPE)
+		g_string_append (str, "JBD2_FLAG_ESCAPE ");
 
-	if (flags & JFS_FLAG_SAME_UUID)
-		g_string_append (str, "JFS_FLAG_SAME_UUID ");
+	if (flags & JBD2_FLAG_SAME_UUID)
+		g_string_append (str, "JBD2_FLAG_SAME_UUID ");
 
-	if (flags & JFS_FLAG_DELETED)
-		g_string_append (str, "JFS_FLAG_DELETED ");
+	if (flags & JBD2_FLAG_DELETED)
+		g_string_append (str, "JBD2_FLAG_DELETED ");
 
-	if (flags & JFS_FLAG_LAST_TAG)
-		g_string_append (str, "JFS_FLAG_LAST_TAG");
+	if (flags & JBD2_FLAG_LAST_TAG)
+		g_string_append (str, "JBD2_FLAG_LAST_TAG");
 
 done:
 	return ;
@@ -325,30 +307,37 @@ void close_pager(FILE *stream)
  * inodestr_to_inode()
  *
  * Returns ino if string is of the form <ino>
- *
- * Copyright (C) 1993, 1994 Theodore Ts'o.  This file may be
- * redistributed under the terms of the GNU Public License.
  */
 int inodestr_to_inode(char *str, uint64_t *blkno)
 {
 	int len;
+	char *buf = NULL;
 	char *end;
+	int ret = OCFS2_ET_INVALID_LOCKRES;
 
 	len = strlen(str);
-	if ((len > 2) && (str[0] == '<') && (str[len-1] == '>')) {
-		if (ocfs2_get_lock_type(str[1]) < OCFS2_NUM_LOCK_TYPES) {
-			if (!ocfs2_decode_lockres(str+1, len-2, NULL, blkno,
-						  NULL))
-				return 0;
-			else
-				return -1;
-		}
-		*blkno = strtoull(str+1, &end, 0);
-		if (*end=='>')
-			return 0;
+	if (!((len > 2) && (str[0] == '<') && (str[len - 1] == '>')))
+		goto bail;
+
+	ret = OCFS2_ET_NO_MEMORY;
+	buf = strndup(str + 1, len - 2);
+	if (!buf)
+		goto bail;
+
+	ret = 0;
+	if (ocfs2_get_lock_type(buf[0]) < OCFS2_NUM_LOCK_TYPES)
+		ret = ocfs2_decode_lockres(buf, NULL, blkno, NULL, NULL);
+	else {
+		*blkno = strtoull(buf, &end, 0);
+		if (*end)
+			ret = OCFS2_ET_INVALID_LOCKRES;
 	}
 
-	return -1;
+bail:
+	if (buf)
+		free(buf);
+
+	return ret;
 }
 
 /*
@@ -736,7 +725,7 @@ errcode_t rdump_inode(ocfs2_filesys *fs, uint64_t blkno, const char *name,
 	} else if (S_ISREG(di->i_mode)) {
 		if (verbose)
 			fprintf(stdout, "%s\n", fullname);
-		fd = open(fullname, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
+		fd = open64(fullname, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
 		if (fd == -1) {
 			com_err(gbls.cmd, errno, "while opening file %s",
 				fullname);
@@ -922,3 +911,68 @@ errcode_t open_debugfs_file(const char *debugfs_path, const char *dirname,
 out:
 	return ret;
 }
+
+void init_stringlist(struct list_head *strlist)
+{
+	INIT_LIST_HEAD(strlist);
+}
+
+errcode_t add_to_stringlist(char *str, struct list_head *strlist)
+{
+	struct strings *s;
+
+	if (!str || !strlen(str))
+		return 0;
+
+	s = calloc(1, sizeof(struct strings));
+	if (!s)
+		return OCFS2_ET_NO_MEMORY;
+
+	INIT_LIST_HEAD(&s->s_list);
+	s->s_str = strdup(str);
+	if (!s->s_str) {
+		free(s);
+		return OCFS2_ET_NO_MEMORY;
+	}
+
+	list_add_tail(&s->s_list, strlist);
+
+	return 0;
+}
+
+void free_stringlist(struct list_head *strlist)
+{
+	struct strings *s;
+	struct list_head *iter, *iter2;
+
+	if (list_empty(strlist))
+		return;
+
+	list_for_each_safe(iter, iter2, strlist) {
+		s = list_entry(iter, struct strings, s_list);
+		list_del(iter);
+		free(s->s_str);
+		free(s);
+	}
+}
+
+int del_from_stringlist(char *str, struct list_head *strlist)
+{
+	struct strings *s;
+	struct list_head *iter, *iter2;
+
+	if (!list_empty(strlist)) {
+		list_for_each_safe(iter, iter2, strlist) {
+			s = list_entry(iter, struct strings, s_list);
+			if (!strcmp(str, s->s_str)) {
+				list_del(iter);
+				free(s->s_str);
+				free(s);
+				return 1;
+			}
+		}
+	}
+
+	return 0;
+}
+
