@@ -66,6 +66,7 @@
 #include "pass4.h"
 #include "problem.h"
 #include "util.h"
+#include "slot_recovery.h"
 
 int verbose = 0;
 
@@ -73,6 +74,8 @@ static char *whoami = "fsck.ocfs2";
 static o2fsck_state _ost;
 static int cluster_locked = 0;
 
+static void mark_magical_clusters(o2fsck_state *ost);
+
 static void handle_signal(int sig)
 {
 	switch (sig) {
@@ -194,6 +197,38 @@ static errcode_t o2fsck_state_init(ocfs2_filesys *fs, o2fsck_state *ost)
 	return 0;
 }
 
+errcode_t o2fsck_state_reinit(ocfs2_filesys *fs, o2fsck_state *ost)
+{
+	errcode_t ret;
+
+	ocfs2_bitmap_free(ost->ost_bad_inodes);
+	ost->ost_bad_inodes = NULL;
+
+	ocfs2_bitmap_free(ost->ost_dir_inodes);
+	ost->ost_dir_inodes = NULL;
+
+	ocfs2_bitmap_free(ost->ost_reg_inodes);
+	ost->ost_reg_inodes = NULL;
+
+	ocfs2_bitmap_free(ost->ost_allocated_clusters);
+	ost->ost_allocated_clusters = NULL;
+
+	o2fsck_icount_free(ost->ost_icount_in_inodes);
+	ost->ost_icount_in_inodes = NULL;
+
+	o2fsck_icount_free(ost->ost_icount_refs);
+	ost->ost_icount_refs = NULL;
+
+	ret = o2fsck_state_init(fs, ost);
+	if (ret) {
+		com_err(whoami, ret, "while intializing o2fsck_state.");
+		return ret;
+	}
+
+	mark_magical_clusters(ost);
+	return 0;
+}
+
 static errcode_t check_superblock(o2fsck_state *ost)
 {
 	struct ocfs2_dinode *di = ost->ost_fs->fs_super;
@@ -415,12 +450,15 @@ static errcode_t maybe_replay_journals(o2fsck_state *ost, char *filename,
 				       int open_flags, uint64_t blkno,
 				       uint64_t blksize)
 {	
-	int replayed = 0, should = 0;
+	int replayed = 0, should = 0, has_dirty = 0;
 	errcode_t ret = 0;
 
-	ret = o2fsck_should_replay_journals(ost->ost_fs, &should);
+	ret = o2fsck_should_replay_journals(ost->ost_fs, &should, &has_dirty);
 	if (ret)
 		goto out;
+
+	ost->ost_has_journal_dirty = has_dirty ? 1 : 0;
+
 	if (!should)
 		goto out;
 
@@ -458,6 +496,43 @@ out:
 	return ret;
 }
 
+/*
+ * Do the slot recovery, replay truncate log, local aloc and orphan dir.
+ * If there is any error, a force check is enabled.
+ */
+static errcode_t o2fsck_slot_recovery(o2fsck_state *ost)
+{
+	errcode_t ret = 0;
+
+	if (!(ost->ost_fs->fs_flags & OCFS2_FLAG_RW)) {
+		printf("** Skipping slot recovery because -n was "
+		       "given. **\n");
+		goto out;
+	}
+
+	ret = o2fsck_replay_local_allocs(ost->ost_fs);
+	if (ret)
+		goto out;
+
+	ret = o2fsck_replay_truncate_logs(ost->ost_fs);
+	if (ret)
+		goto out;
+
+	/*
+	 * If the user want a force-check, orphan_dir will be
+	 * replayed after the full check.
+	 */
+	if (!ost->ost_force) {
+		ret = o2fsck_replay_orphan_dirs(ost);
+		if (ret)
+			com_err(whoami, ret, "while trying to replay the orphan"
+				" directory");
+	}
+
+out:
+	return ret;
+}
+
 static errcode_t recover_backup_super(o2fsck_state *ost,
 				      char* device, int sb_num)
 {
@@ -575,6 +650,7 @@ int main(int argc, char **argv)
 	int c, open_flags = OCFS2_FLAG_RW | OCFS2_FLAG_STRICT_COMPAT_CHECK;
 	int sb_num = 0;
 	int fsck_mask = FSCK_OK;
+	int slot_recover_err = 0;
 	errcode_t ret;
 
 	memset(ost, 0, sizeof(o2fsck_state));
@@ -786,9 +862,17 @@ int main(int argc, char **argv)
 		goto unlock;
 	}
 
+	ret = o2fsck_slot_recovery(ost);
+	if (ret) {
+		printf("fsck encountered errors while recovering slot "
+		       "information, check forced.\n");
+		slot_recover_err = 1;
+		ost->ost_force = 1;
+	}
+
 	if (fs_is_clean(ost, filename)) {
 		fsck_mask = FSCK_OK;
-		goto unlock;
+		goto clear_dirty_flag;
 	}
 
 #if 0
@@ -837,15 +921,39 @@ done:
 	if (ret)
 		fsck_mask |= FSCK_ERROR;
 	else {
+		fsck_mask = FSCK_OK;
 		ost->ost_saw_error = 0;
 		printf("All passes succeeded.\n");
 	}
 
+clear_dirty_flag:
 	if (ost->ost_fs->fs_flags & OCFS2_FLAG_RW) {
 		ret = write_out_superblock(ost);
 		if (ret)
 			com_err(whoami, ret, "while writing back the "
 				"superblock(s)");
+		if (fsck_mask == FSCK_OK) {
+			if (slot_recover_err) {
+				ret = o2fsck_slot_recovery(ost);
+				if (ret) {
+					com_err(whoami, ret, "while doing slot "
+						"recovery.");
+					goto unlock;
+				}
+			}
+
+			ret = o2fsck_clear_journal_flags(ost);
+			if (ret) {
+				com_err(whoami, ret, "while clear dirty "
+					"journal flag.");
+				goto unlock;
+			}
+
+			ret = ocfs2_format_slot_map(ost->ost_fs);
+			if (ret)
+				com_err(whoami, ret, "while format slot "
+					"map.");
+		}
 	}
 
 unlock:
