@@ -70,9 +70,13 @@ typedef struct _o2fsck_state {
  					  * errors as we exit */
  			ost_stale_mounts:1, /* set when reading publish blocks
  					     * that still indicated mounted */
-			ost_fix_fs_gen:1;
+			ost_fix_fs_gen:1,
+			ost_has_journal_dirty:1;
+	errcode_t ost_err;
 } o2fsck_state;
 
+errcode_t o2fsck_state_reinit(ocfs2_filesys *fs, o2fsck_state *ost);
+
 /* The idea is to let someone off-site run fsck and have it give us 
  * enough information to diagnose problems with */
 extern int verbose;
