@@ -3,30 +3,25 @@
  *
  * Copyright (C) 2000 Andreas Dilger
  * Copyright (C) 2000 Theodore Ts'o
- * Copyright (C) 2004 Oracle.  All rights reserved.
+ * Copyright (C) 2004,2008 Oracle.  All rights reserved.
  *
  * Parts of the code are based on fs/jfs/journal.c by Stephen C. Tweedie
  * Copyright (C) 1999 Red Hat Software
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
- * License, version 2,  as published by the Free Software Foundation.
+ * License version 2 as published by the Free Software Foundation.
  * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
  * 
- * You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 021110-1307, USA.
- *
  * --
- * This replays the jbd journals for each slot.  First all the journals are
+ * This replays the jbd2 journals for each slot.  First all the journals are
  * walked to detect inconsistencies.  Only journals with no problems will be
  * replayed.  IO errors during replay will just result in partial journal
- * replay, just like jbd does in the kernel.  Journals that don't pass
+ * replay, just like jbd2 does in the kernel.  Journals that don't pass
  * consistency checks, like having overlapping blocks or strange fields, are
  * ignored and left for later passes to clean up.  
 
@@ -202,29 +197,31 @@ static uint64_t jwrap(journal_superblock_t *jsb, uint64_t block)
 	return block;
 }
 
-static errcode_t count_tags(ocfs2_filesys *fs, char *buf, size_t size,
-			    uint64_t *nr_ret)
+static errcode_t count_tags(ocfs2_filesys *fs, journal_superblock_t *jsb,
+			    char *buf, uint64_t *nr_ret)
 {
-	journal_block_tag_t *tag, *last;
+	char *tagp, *last;
+	journal_block_tag_t *tag;
+	int tag_bytes = ocfs2_journal_tag_bytes(jsb);
 	uint64_t nr = 0;
 
-	if (size < sizeof(journal_header_t) + sizeof(*tag))
+	if (jsb->s_blocksize < sizeof(journal_header_t) + tag_bytes)
 		return OCFS2_ET_BAD_JOURNAL_TAG;
 
-       	tag = (journal_block_tag_t *)&buf[sizeof(journal_header_t)];
-       	last = (journal_block_tag_t *)&buf[size - sizeof(*tag)];
+	tagp = &buf[sizeof(journal_header_t)];
+	last = &buf[jsb->s_blocksize - tag_bytes];
 
-	for(; tag <= last; tag++) {
+	for(; tagp <= last; tagp += tag_bytes) {
+		tag = (journal_block_tag_t *)tagp;
 		nr++;
 		if (ocfs2_block_out_of_range(fs, 
-					     be32_to_cpu(tag->t_blocknr)))
+					     ocfs2_journal_tag_block(tag, tag_bytes)))
 			return OCFS2_ET_BAD_JOURNAL_TAG;
 
-		if (tag->t_flags & cpu_to_be32(JFS_FLAG_LAST_TAG))
+		if (tag->t_flags & cpu_to_be32(JBD2_FLAG_LAST_TAG))
 			break;
-		/* inline uuids are 16 bytes, tags are 8 */
-		if (!(tag->t_flags & cpu_to_be32(JFS_FLAG_SAME_UUID)))
-			tag += 2;
+		if (!(tag->t_flags & cpu_to_be32(JBD2_FLAG_SAME_UUID)))
+			tagp += 16;
 	}
 
 	*nr_ret = nr;
@@ -288,14 +285,18 @@ static errcode_t read_journal_block(ocfs2_filesys *fs,
 static errcode_t replay_blocks(ocfs2_filesys *fs, struct journal_info *ji,
 			       char *buf, uint64_t seq, uint64_t *next_block)
 {
-	journal_block_tag_t tag, *tagp;
+	char *tagp;
+	journal_block_tag_t *tag;
 	size_t i, num;
 	char *io_buf = NULL;
 	errcode_t err, ret = 0;
+	int tag_bytes = ocfs2_journal_tag_bytes(ji->ji_jsb);
+	uint32_t t_flags;
+	uint64_t block64;
 		
-	tagp = (journal_block_tag_t *)(buf + sizeof(journal_header_t));
+	tagp = buf + sizeof(journal_header_t);
 	num = (ji->ji_jsb->s_blocksize - sizeof(journal_header_t)) / 
-	      sizeof(tag);
+		tag_bytes;
 
 	ret = ocfs2_malloc_blocks(fs->fs_io, 1, &io_buf);
 	if (ret) {
@@ -303,17 +304,17 @@ static errcode_t replay_blocks(ocfs2_filesys *fs, struct journal_info *ji,
 		goto out;
 	}
 
-	for(i = 0; i < num; i++, tagp++, (*next_block)++) {
-		memcpy(&tag, tagp, sizeof(tag));
-		tag.t_flags = be32_to_cpu(tag.t_flags);
-		tag.t_blocknr = be32_to_cpu(tag.t_blocknr);
+	for(i = 0; i < num; i++, tagp += tag_bytes, (*next_block)++) {
+		tag = (journal_block_tag_t *)tagp;
+		t_flags = be32_to_cpu(tag->t_flags);
+		block64 = ocfs2_journal_tag_block(tag, tag_bytes);
 
 		*next_block = jwrap(ji->ji_jsb, *next_block);
 
 		verbosef("recovering journal block %"PRIu64" to disk block "
-			 "%"PRIu32"\n", *next_block, tag.t_blocknr);	
+			 "%"PRIu64"\n", *next_block, block64);
 
-		if (revoke_this_block(&ji->ji_revoke, tag.t_blocknr, seq))
+		if (revoke_this_block(&ji->ji_revoke, block64, seq))
 			goto skip_io;
 
 		err = read_journal_block(fs, ji, *next_block, io_buf, 1);
@@ -322,22 +323,20 @@ static errcode_t replay_blocks(ocfs2_filesys *fs, struct journal_info *ji,
 			goto skip_io;
 		}
 
-		if (tag.t_flags & JFS_FLAG_ESCAPE) {
-			uint32_t magic = cpu_to_be32(JFS_MAGIC_NUMBER);
+		if (t_flags & JBD2_FLAG_ESCAPE) {
+			uint32_t magic = cpu_to_be32(JBD2_MAGIC_NUMBER);
 			memcpy(io_buf, &magic, sizeof(magic));
 		}
 
-		err = io_write_block(fs->fs_io, tag.t_blocknr, 1, 
-				     io_buf);
+		err = io_write_block(fs->fs_io, block64, 1, io_buf);
 		if (err)
 			ret = err;
 
 	skip_io:
-		if (tag.t_flags & JFS_FLAG_LAST_TAG)
+		if (t_flags & JBD2_FLAG_LAST_TAG)
 			i = num; /* be sure to increment next_block */
-		/* inline uuids are 16 bytes, tags are 8 */
-		if (!(tag.t_flags & JFS_FLAG_SAME_UUID))
-			tagp += 2;
+		if (!(t_flags & JBD2_FLAG_SAME_UUID))
+			tagp += 16;
 	}
 	
 out:
@@ -389,7 +388,7 @@ static errcode_t walk_journal(ocfs2_filesys *fs, int slot,
 
 		verbosef("jh magic %x\n", jh.h_magic);
 
-		if (jh.h_magic != JFS_MAGIC_NUMBER)
+		if (jh.h_magic != JBD2_MAGIC_NUMBER)
 			break;
 
 		verbosef("jh block %x\n", jh.h_blocktype);
@@ -399,7 +398,7 @@ static errcode_t walk_journal(ocfs2_filesys *fs, int slot,
 			break;
 
 		switch(jh.h_blocktype) {
-		case JFS_DESCRIPTOR_BLOCK:
+		case JBD2_DESCRIPTOR_BLOCK:
 			verbosef("found a desc type %x\n", jh.h_blocktype);
 			/* replay the blocks described in the desc block */
 			if (recover) {
@@ -411,19 +410,19 @@ static errcode_t walk_journal(ocfs2_filesys *fs, int slot,
 			}
 
 			/* just record the blocks as used and carry on */ 
-			err = count_tags(fs, buf, jsb->s_blocksize, &nr);
+			err = count_tags(fs, jsb, buf, &nr);
 			if (err)
 				ret = err;
 			else
 				next_block = jwrap(jsb, next_block + nr);
 			break;
 
-		case JFS_COMMIT_BLOCK:
+		case JBD2_COMMIT_BLOCK:
 			verbosef("found a commit type %x\n", jh.h_blocktype);
 			next_seq++;
 			break;
 
-		case JFS_REVOKE_BLOCK:
+		case JBD2_REVOKE_BLOCK:
 			verbosef("found a revoke type %x\n", jh.h_blocktype);
 			add_revoke_records(ji, buf, jsb->s_blocksize,
 					   next_seq);
@@ -516,7 +515,8 @@ out:
  * stop a second journal replay but leave the dirty bit set so that the kernel
  * will truncate the orphaned inodes. 
  */
-errcode_t o2fsck_should_replay_journals(ocfs2_filesys *fs, int *should)
+errcode_t o2fsck_should_replay_journals(ocfs2_filesys *fs, int *should,
+					int *has_dirty)
 {
 	uint16_t i, max_slots;
 	char *buf = NULL;
@@ -564,6 +564,8 @@ errcode_t o2fsck_should_replay_journals(ocfs2_filesys *fs, int *should)
 		verbosef("slot %d JOURNAL_DIRTY_FL: %d\n", i, is_dirty);
 		if (!is_dirty)
 			continue;
+		else
+			*has_dirty = 1;
 
 		ret = ocfs2_extent_map_get_blocks(cinode, 0, 1, &blkno,
 						  &contig, NULL);
@@ -725,38 +727,187 @@ out:
 	return ret;
 }
 
-static errcode_t check_journal_super(ocfs2_filesys *fs,
-				     ocfs2_cached_inode *ci)
+struct journal_check_info {
+	errcode_t		i_error;
+	uint32_t		i_clusters;
+	ocfs2_fs_options	i_features;
+};
+
+struct journal_check_context {
+	int				jc_max_slots;
+	int				jc_this_slot;
+	uint32_t			jc_max_clusters; /* Size of the
+							    largest journal
+							    found */
+	ocfs2_fs_options		jc_max_features; /* Union all
+							    features in
+							    good journals */
+	struct journal_check_info	*jc_info;	/* One per slot */
+};
+
+static errcode_t check_journals_func(o2fsck_state *ost,
+				     ocfs2_cached_inode *ci,
+				     struct journal_check_context *jc)
 {
-	errcode_t ret;
+	errcode_t err, ret;
+	ocfs2_filesys *fs = ost->ost_fs;
 	uint64_t contig;
 	uint64_t blkno;
 	char *buf = NULL;
+	journal_superblock_t *jsb;
+	struct journal_check_info *ji = &(jc->jc_info[jc->jc_this_slot]);
 
 	ret = ocfs2_malloc_blocks(fs->fs_io, 1, &buf);
 	if (ret)
 		goto out;
 
-	ret = ocfs2_extent_map_get_blocks(ci, 0, 1, &blkno, &contig, NULL);
-	if (ret)
+	err = ocfs2_extent_map_get_blocks(ci, 0, 1, &blkno, &contig, NULL);
+	if (err) {
+		ji->i_error = err;
 		goto out;
+	}
+
+	ji->i_clusters = ci->ci_inode->i_clusters;
+
+	err = ocfs2_read_journal_superblock(fs, blkno, buf);
+	if (err) {
+		ji->i_error = err;
+		goto out;
+	}
+
+	jsb = (journal_superblock_t *)buf;
+	ji->i_features.opt_compat = jsb->s_feature_compat;
+	ji->i_features.opt_ro_compat = jsb->s_feature_ro_compat;
+	ji->i_features.opt_incompat = jsb->s_feature_incompat;
+
+	if (!ji->i_clusters) {
+		ji->i_error = OCFS2_ET_JOURNAL_TOO_SMALL;
+		goto out;
+	}
+
+	jc->jc_max_clusters = ocfs2_max(jc->jc_max_clusters,
+					ci->ci_inode->i_clusters);
+	jc->jc_max_features.opt_compat |= jsb->s_feature_compat;
+	jc->jc_max_features.opt_ro_compat |= jsb->s_feature_ro_compat;
+	jc->jc_max_features.opt_incompat |= jsb->s_feature_incompat;
+	ji->i_error = 0;
+
+out:
+	if (buf)
+		ocfs2_free(&buf);
+
+	return ret;
+}
+
+static errcode_t fix_journals_func(o2fsck_state *ost,
+				   ocfs2_cached_inode *ci,
+				   struct journal_check_context *jc)
+{
+	errcode_t err, ret = 0;
+	ocfs2_filesys *fs = ost->ost_fs;
+	char fname[OCFS2_MAX_FILENAME_LEN];
+	struct journal_check_info *ji = &(jc->jc_info[jc->jc_this_slot]);
+	uint32_t min_clusters =
+			ocfs2_clusters_in_bytes(fs,
+						OCFS2_MIN_JOURNAL_SIZE);
+
+	ocfs2_sprintf_system_inode_name(fname, OCFS2_MAX_FILENAME_LEN,
+					JOURNAL_SYSTEM_INODE,
+					jc->jc_this_slot);
+
+	if (ji->i_error &&
+	    (ji->i_error != OCFS2_ET_JOURNAL_TOO_SMALL) &&
+	    (ji->i_error != OCFS2_ET_UNSUPP_FEATURE) &&
+	    (ji->i_error != OCFS2_ET_RO_UNSUPP_FEATURE)) {
+		if (prompt(ost, PY, PR_JOURNAL_FILE_INVALID,
+			    "journal file %s is invalid, regenerate it?",
+			    fname)) {
+			err = ocfs2_make_journal(fs, ci->ci_blkno,
+						 jc->jc_max_clusters,
+						 &jc->jc_max_features);
+
+			ji->i_error = err;
+		}
+		goto out;
+	}
+
+	if ((ji->i_error == OCFS2_ET_UNSUPP_FEATURE) ||
+	    (ji->i_error == OCFS2_ET_RO_UNSUPP_FEATURE)) {
+		if (prompt(ost, PN, PR_JOURNAL_FILE_INVALID,
+			   "journal file %s has unknown features.  "
+			   "However, other journals have only known "
+			   "features, so this is likely a corruption.  "
+			   "If you think your filesystem may be newer "
+			   "than this version of fsck.ocfs2, say N here "
+			   "and grab the latest version of fsck.ocfs2.  "
+			   "Reset the journal features to match other "
+			   "journals?",
+			   fname)) {
+			err = ocfs2_make_journal(fs, ci->ci_blkno,
+						 ocfs2_max(ji->i_clusters,
+							   min_clusters),
+						 &jc->jc_max_features);
+			if (!err) {
+				ji->i_features.opt_compat =
+					jc->jc_max_features.opt_compat;
+				ji->i_features.opt_ro_compat =
+					jc->jc_max_features.opt_ro_compat;
+				ji->i_features.opt_incompat =
+					jc->jc_max_features.opt_incompat;
+			}
+			ji->i_error = err;
+		}
+	} else if ((ji->i_features.opt_compat !=
+		    jc->jc_max_features.opt_compat) ||
+		   (ji->i_features.opt_ro_compat !=
+		    jc->jc_max_features.opt_ro_compat) ||
+		   (ji->i_features.opt_incompat !=
+		    jc->jc_max_features.opt_incompat)) {
+		if (prompt(ost, PY, PR_JOURNAL_FILE_INVALID,
+			   "journal file %s is missing features that "
+			   "are set on other journal files.  Set these "
+			   "features?",
+			   fname)) {
+			err = ocfs2_make_journal(fs, ci->ci_blkno,
+						 ocfs2_max(ji->i_clusters,
+							   min_clusters),
+						 &jc->jc_max_features);
+			if (!err) {
+				ji->i_features.opt_compat =
+					jc->jc_max_features.opt_compat;
+				ji->i_features.opt_ro_compat =
+					jc->jc_max_features.opt_ro_compat;
+				ji->i_features.opt_incompat =
+					jc->jc_max_features.opt_incompat;
+			}
+			ji->i_error = err;
+		}
+	}
+
+	if (ji->i_clusters != jc->jc_max_clusters) {
+		if (prompt(ost, PY, PR_JOURNAL_FILE_INVALID,
+			   "journal file %s is too small, extend it?",
+			   fname)) {
+			err = ocfs2_make_journal(fs, ci->ci_blkno,
+						 jc->jc_max_clusters,
+						 &ji->i_features);
+			ji->i_error = err;
+		}
+	}
 
-	ret = ocfs2_read_journal_superblock(fs, blkno, buf);
 out:
 	return ret;
 }
 
-/* When we remove slot in tunefs.ocfs2, there may be some panic and
- * we may corrupt some journal files, so we have to check whether the
- * journal file is corrupted and recreate it.
- */
-errcode_t o2fsck_check_journals(o2fsck_state *ost)
+static errcode_t check_journal_walk(o2fsck_state *ost,
+				    errcode_t (*func)(o2fsck_state *ost,
+						      ocfs2_cached_inode *ci,
+						      struct journal_check_context *jc),
+				    struct journal_check_context *jc)
 {
 	errcode_t ret = 0;
 	uint64_t blkno;
-	uint32_t num_clusters = 0;
 	ocfs2_filesys *fs = ost->ost_fs;
-	char fname[OCFS2_MAX_FILENAME_LEN];
 	uint16_t i, max_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
 	ocfs2_cached_inode *ci = NULL;
 
@@ -764,43 +915,162 @@ errcode_t o2fsck_check_journals(o2fsck_state *ost)
 		ret = ocfs2_lookup_system_inode(fs, JOURNAL_SYSTEM_INODE, i,
 						&blkno);
 		if (ret)
-			goto out;
+			break;
 
 		ret = ocfs2_read_cached_inode(fs, blkno, &ci);
 		if (ret)
-			goto out;
+			break;
 
-		if (ci->ci_inode->i_clusters > 0) {
-			/* check whether the file contains valid super block. */
-			ret = check_journal_super(fs, ci);
-			if (!ret) {
-				/* record the valid cluster size. */
-				num_clusters = ci->ci_inode->i_clusters;
-				continue;
-			}
-		}
-
-		if (num_clusters == 0) {
-			/* none of the journal is valid, servere errors. */
-			ret = OCFS2_ET_JOURNAL_TOO_SMALL;
-			goto out;
-		}
-
-		sprintf(fname,
-			ocfs2_system_inodes[JOURNAL_SYSTEM_INODE].si_name, i);
-		if (!prompt(ost, PY, PR_JOURNAL_FILE_INVALID,
-			    "journal file %s is invalid, regenerate it?",
-			    fname))
-			continue;
-
-		ret = ocfs2_make_journal(fs, blkno, num_clusters);
+		jc->jc_this_slot = i;
+		ret = func(ost, ci, jc);
 		if (ret)
-			goto out;
+			break;
 	}
 
-out:
 	if (ci)
 		ocfs2_free_cached_inode(fs, ci);
+
 	return ret;
 }
 
+errcode_t o2fsck_check_journals(o2fsck_state *ost)
+{
+	errcode_t ret;
+	int i;
+	ocfs2_filesys *fs = ost->ost_fs;
+	int have_one_good_journal = 0, problem_is_consistent = 1;
+	errcode_t known_problem = 0;
+	struct journal_check_context jc = {
+		.jc_max_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots,
+	};
+	struct journal_check_info *ji;
+
+	ret = ocfs2_malloc0(sizeof(struct journal_check_info) *
+			    jc.jc_max_slots,
+			    &jc.jc_info);
+	if (ret) {
+		com_err(whoami, ret, "while checking journals");
+		goto out;
+	}
+
+	ret = check_journal_walk(ost, check_journals_func, &jc);
+	if (ret) {
+		com_err(whoami, ret, "while checking journals");
+		goto out;
+	}
+
+	/*
+	 * We now know the state of all our journals.  If we have at least
+	 * one good journal, we have a sane state to fix the others from.
+	 * We require all our journals to have identical configuration.
+	 * Any inconsistencies (invalid size, bad feature flags) are
+	 * probably corruption or a failed tunefs.
+	 *
+	 * If we don't have a good journal, but all the journals have the
+	 * exact same problem, we may be able to handle it as well.  We
+	 * currently know how to handle these problems:
+	 *
+	 * JOURNAL_TOO_SMALL
+	 *
+	 * We simply allocate a default journal size.
+	 *
+	 * UNSUPP_FEATURE & RO_UNSUPP_FEATURE
+	 *
+	 * If one journal has an unsupported feature bit set, it's probably
+	 * corruption.  If all the journals have the exact same feature
+	 * bit set, it's certainly a feature we don't understand, and we
+	 * want the user to upgrade their fsck.
+	 */
+	for (i = 0; i < jc.jc_max_slots; i++) {
+		ji = &jc.jc_info[i];
+
+		if (!ji->i_error) {
+			have_one_good_journal = 1;
+			continue;
+		}
+
+		if ((ji->i_error != OCFS2_ET_JOURNAL_TOO_SMALL) &&
+		    (ji->i_error != OCFS2_ET_UNSUPP_FEATURE) &&
+		    (ji->i_error != OCFS2_ET_RO_UNSUPP_FEATURE)) {
+			problem_is_consistent = 0;
+			continue;
+		}
+
+		if (known_problem) {
+			if (known_problem != ji->i_error)
+				problem_is_consistent = 0;
+			continue;
+		}
+
+		known_problem = ji->i_error;
+	}
+
+	if (!have_one_good_journal) {
+		if (!problem_is_consistent || !known_problem) {
+			ret = jc.jc_info[0].i_error;
+			com_err(whoami, ret, "while checking journals");
+			goto out;
+		}
+		if ((known_problem == OCFS2_ET_UNSUPP_FEATURE) ||
+		    (known_problem == OCFS2_ET_RO_UNSUPP_FEATURE)) {
+			com_err(whoami, known_problem,
+				"on all journals.  Please upgrade to the "
+				"latest version of fsck.ocfs2");
+			ret = known_problem;
+			goto out;
+		}
+
+		if (known_problem != OCFS2_ET_JOURNAL_TOO_SMALL) {
+			ret = known_problem;
+			com_err(whoami, ret, "for all journals");
+			goto out;
+		}
+
+		/* Force a valid cluster count for the journals */
+		jc.jc_max_clusters =
+			ocfs2_clusters_in_bytes(fs,
+						OCFS2_MIN_JOURNAL_SIZE);
+	}
+
+	ret = check_journal_walk(ost, fix_journals_func, &jc);
+
+out:
+	if (jc.jc_info)
+		ocfs2_free(&jc.jc_info);
+
+	return ret;
+}
+
+static errcode_t ocfs2_clear_journal_flag(ocfs2_filesys *fs,
+					  struct ocfs2_dinode *di,
+					  int slot)
+{
+	errcode_t ret = 0;
+
+	if (!(di->i_flags & OCFS2_VALID_FL) ||
+	    !(di->i_flags & OCFS2_SYSTEM_FL) ||
+	    !(di->i_flags & OCFS2_JOURNAL_FL))
+		return OCFS2_ET_INVALID_ARGUMENT;
+
+	if (!(di->id1.journal1.ij_flags & OCFS2_JOURNAL_DIRTY_FL))
+		goto bail;
+
+	di->id1.journal1.ij_flags &= ~OCFS2_JOURNAL_DIRTY_FL;
+
+	ret = ocfs2_write_inode(fs, di->i_blkno, (char *)di);
+	if (!ret)
+		printf("Slot %d's journal dirty flag removed\n", slot);
+
+bail:
+	return ret;
+}
+
+errcode_t o2fsck_clear_journal_flags(o2fsck_state *ost)
+{
+	if (!ost->ost_has_journal_dirty)
+		return 0;
+
+	return handle_slots_system_file(ost->ost_fs,
+					JOURNAL_SYSTEM_INODE,
+					ocfs2_clear_journal_flag);
+}
