@@ -83,12 +83,19 @@ static int dirent_has_dots(struct ocfs2_dir_entry *dirent, int num_dots)
 	return dirent->name[0] == '.';
 }
 
-static int expected_dots(o2fsck_dirblock_entry *dbe, int offset)
+static int expected_dots(o2fsck_state *ost,
+			 o2fsck_dirblock_entry *dbe,
+			 int offset)
 {
+	int inline_off = offsetof(struct ocfs2_dinode, id2.i_data.id_data);
+
 	if (dbe->e_blkcount == 0) {
-		if (offset == 0)
+		if (offset == 0 ||
+		    (dbe->e_ino == dbe->e_blkno && offset == inline_off))
 			return 1;
-		if (offset == OCFS2_DIR_REC_LEN(1))
+		if (offset == OCFS2_DIR_REC_LEN(1) ||
+		    (dbe->e_ino == dbe->e_blkno &&
+		     offset == inline_off + OCFS2_DIR_REC_LEN(1)))
 			return 2;
 	}
 
@@ -99,7 +106,7 @@ static errcode_t fix_dirent_dots(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
 				 struct ocfs2_dir_entry *dirent, int offset, 
 				 int left, unsigned int *flags)
 {
-	int expect_dots = expected_dots(dbe, offset);
+	int expect_dots = expected_dots(ost, dbe, offset);
 	int changed_len = 0;
 	struct ocfs2_dir_entry *next;
 	uint16_t new_len;
@@ -186,6 +193,69 @@ out:
 	return ret;
 }
 
+/*
+ * The directory trailer has compatibility fields so it can be treated
+ * as an empty (deleted) dirent.  We need to make sure those are correct.
+ */
+static void fix_dir_trailer(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
+			    struct ocfs2_dir_block_trailer *trailer,
+			    unsigned int *flags)
+{
+	if (trailer->db_compat_inode &&
+	    prompt(ost, PY, PR_DIR_TRAILER_INODE,
+		   "Directory block trailer for logical block %"PRIu64" "
+		   "physcal block %"PRIu64" in directory inode %"PRIu64" "
+		   "has a non-zero inode number.  Clear it?",
+		   dbe->e_blkcount, dbe->e_blkno, dbe->e_ino)) {
+		trailer->db_compat_inode = 0;
+		*flags |= OCFS2_DIRENT_CHANGED;
+	}
+
+	if (trailer->db_compat_name_len &&
+	    prompt(ost, PY, PR_DIR_TRAILER_NAME_LEN,
+		   "Directory block trailer for logical block %"PRIu64" "
+		   "physcal block %"PRIu64" in directory inode %"PRIu64" "
+		   "has a non-zero name_len.  Clear it?",
+		   dbe->e_blkcount, dbe->e_blkno, dbe->e_ino)) {
+		trailer->db_compat_name_len = 0;
+		*flags |= OCFS2_DIRENT_CHANGED;
+	}
+
+	if ((trailer->db_compat_rec_len !=
+	     sizeof(struct ocfs2_dir_block_trailer)) &&
+	    prompt(ost, PY, PR_DIR_TRAILER_REC_LEN,
+		   "Directory block trailer for logical block %"PRIu64" "
+		   "physcal block %"PRIu64" in directory inode %"PRIu64" "
+		   "has an invalid rec_len.  Fix it?",
+		   dbe->e_blkcount, dbe->e_blkno, dbe->e_ino)) {
+		trailer->db_compat_rec_len =
+			sizeof(struct ocfs2_dir_block_trailer);
+		*flags |= OCFS2_DIRENT_CHANGED;
+	}
+
+	if ((trailer->db_blkno != dbe->e_blkno) &&
+	    prompt(ost, PY, PR_DIR_TRAILER_BLKNO,
+		   "Directory block trailer for logical block %"PRIu64" "
+		   "physcal block %"PRIu64" in directory inode %"PRIu64" "
+		   "has an invalid db_blkno of %"PRIu64".  Fix it?",
+		   dbe->e_blkcount, dbe->e_blkno, dbe->e_ino,
+		   trailer->db_blkno)) {
+		trailer->db_blkno = dbe->e_blkno;
+		*flags |= OCFS2_DIRENT_CHANGED;
+	}
+
+	if ((trailer->db_parent_dinode != dbe->e_ino) &&
+	    prompt(ost, PY, PR_DIR_TRAILER_PARENT_INODE,
+		   "Directory block trailer for logical block %"PRIu64" "
+		   "physcal block %"PRIu64" in directory inode %"PRIu64" "
+		   "claims it belongs to inoe %"PRIu64".  Fix it?",
+		   dbe->e_blkcount, dbe->e_blkno, dbe->e_ino,
+		   trailer->db_parent_dinode)) {
+		trailer->db_parent_dinode = dbe->e_ino;
+		*flags |= OCFS2_DIRENT_CHANGED;
+	}
+}
+
 static int dirent_leaves_partial(struct ocfs2_dir_entry *dirent, int left)
 {
 	left -= dirent->rec_len;
@@ -432,7 +502,7 @@ static errcode_t fix_dirent_linkage(o2fsck_state *ost,
 				    int offset,
 				    unsigned int *flags)
 {
-	int expect_dots = expected_dots(dbe, offset);
+	int expect_dots = expected_dots(ost, dbe, offset);
 	o2fsck_dir_parent *dp;
 	errcode_t ret = 0;
 	int is_dir;
@@ -605,7 +675,7 @@ static unsigned pass2_dir_block_iterate(o2fsck_dirblock_entry *dbe,
 {
 	struct dirblock_data *dd = priv_data;
 	struct ocfs2_dir_entry *dirent, *prev = NULL;
-	unsigned int offset = 0, ret_flags = 0;
+	unsigned int offset = 0, ret_flags = 0, end = dd->fs->fs_blocksize;
 	struct ocfs2_dinode *di = (struct ocfs2_dinode *)dd->inoblock_buf; 
 	errcode_t ret = 0;
 
@@ -637,17 +707,31 @@ static unsigned pass2_dir_block_iterate(o2fsck_dirblock_entry *dbe,
 	verbosef("dir block %"PRIu64" block offs %"PRIu64" in ino\n",
 		 dbe->e_blkno, dbe->e_blkcount);
 
-	if (dbe->e_blkcount >= ocfs2_blocks_in_bytes(dd->fs, di->i_size))
-		goto out;
+	if (di->i_dyn_features & OCFS2_INLINE_DATA_FL) {
+		if (dbe->e_ino != dbe->e_blkno)
+			goto out;
 
- 	ret = ocfs2_read_dir_block(dd->fs, dbe->e_blkno, dd->dirblock_buf);
-	if (ret && ret != OCFS2_ET_DIR_CORRUPTED) {
-		com_err(whoami, ret, "while reading dir block %"PRIu64,
-			dbe->e_blkno);
-		goto out;
+		memcpy(dd->dirblock_buf, dd->inoblock_buf,
+		       dd->fs->fs_blocksize);
+		offset = offsetof(struct ocfs2_dinode, id2.i_data.id_data);
+	} else {
+		if (dbe->e_blkcount >= ocfs2_blocks_in_bytes(dd->fs,
+							     di->i_size))
+			goto out;
+
+		ret = ocfs2_read_dir_block(dd->fs, di, dbe->e_blkno,
+					   dd->dirblock_buf);
+		if (ret && ret != OCFS2_ET_DIR_CORRUPTED) {
+			com_err(whoami, ret, "while reading dir block %"PRIu64,
+				dbe->e_blkno);
+			goto out;
+		}
+
+		if (ocfs2_dir_has_trailer(dd->fs, di))
+			end = ocfs2_dir_trailer_blk_off(dd->fs);
 	}
 
-	while (offset < dd->fs->fs_blocksize) {
+	while (offset < end) {
 		dirent = (struct ocfs2_dir_entry *)(dd->dirblock_buf + offset);
 
 		verbosef("checking dirent offset %d, rec_len %"PRIu16" "
@@ -661,7 +745,7 @@ static unsigned pass2_dir_block_iterate(o2fsck_dirblock_entry *dbe,
 		/* if we can't trust this dirent then fix it up or skip
 		 * the whole block */
 		if (corrupt_dirent_lengths(dirent,
-					   dd->fs->fs_blocksize - offset)) {
+					   end - offset)) {
 			if (!prompt(dd->ost, PY, PR_DIRENT_LENGTH,
 				    "Directory inode %"PRIu64" "
 				    "corrupted in logical block %"PRIu64" "
@@ -673,8 +757,7 @@ static unsigned pass2_dir_block_iterate(o2fsck_dirblock_entry *dbe,
 
 			/* we edit the dirent in place so we try to parse
 			 * it again after fixing it */
-			fix_dirent_lengths(dirent,
-					   dd->fs->fs_blocksize - offset,
+			fix_dirent_lengths(dirent, end - offset,
 					   prev, &ret_flags);
 			continue;
 
@@ -690,8 +773,7 @@ static unsigned pass2_dir_block_iterate(o2fsck_dirblock_entry *dbe,
 		 * XXX should verify that ocfs2 reclaims entries like that.
 		 */
 		ret = fix_dirent_dots(dd->ost, dbe, dirent, offset, 
-					     dd->fs->fs_blocksize - offset,
-					     &ret_flags);
+				      end - offset, &ret_flags);
 		if (ret)
 			goto out;
 		if (dirent->inode == 0)
@@ -734,9 +816,21 @@ next:
 		prev = dirent;
 	}
 
+	if (ocfs2_dir_has_trailer(dd->fs, di))
+		fix_dir_trailer(dd->ost, dbe,
+				ocfs2_dir_trailer_from_block(dd->fs,
+							     dd->dirblock_buf),
+				&ret_flags);
+
 	if (ret_flags & OCFS2_DIRENT_CHANGED) {
-		ret = ocfs2_write_dir_block(dd->fs, dbe->e_blkno,
-					    dd->dirblock_buf);
+		if (di->i_dyn_features & OCFS2_INLINE_DATA_FL) {
+			memcpy(dd->inoblock_buf, dd->dirblock_buf,
+			       dd->fs->fs_blocksize);
+			ret = ocfs2_write_inode(dd->fs, dbe->e_ino,
+						dd->dirblock_buf);
+		} else
+			ret = ocfs2_write_dir_block(dd->fs, di, dbe->e_blkno,
+						    dd->dirblock_buf);
 		if (ret) {
 			com_err(whoami, ret, "while writing dir block %"PRIu64,
 				dbe->e_blkno);
