@@ -53,13 +53,14 @@
 #include <o2dlm/o2dlm.h>
 #include <o2cb/o2cb.h>
 #include <ocfs2/ocfs2_err.h>
-#include <ocfs2/jbd.h>
+#include <ocfs2/jbd2.h>
 #include <ocfs2-kernel/ocfs2_lockid.h>
 
 #define OCFS2_LIB_FEATURE_INCOMPAT_SUPP		(OCFS2_FEATURE_INCOMPAT_SUPP | \
 						 OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV | \
 						 OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG | \
 						 OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT   | \
+						 OCFS2_FEATURE_INCOMPAT_INLINE_DATA   | \
 						 OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG)
 
 #define OCFS2_LIB_FEATURE_RO_COMPAT_SUPP	OCFS2_FEATURE_RO_COMPAT_SUPP
@@ -68,20 +69,6 @@
 
 #define OCFS2_LIB_ABORTED_TUNEFS_SUPP		OCFS2_TUNEFS_INPROG_REMOVE_SLOT
 
-/*
- * Printable strings for feature flags
- * These MUST be kept in sync with the flags in ocfs2_fs.h, and MUST
- * be named <flag>_STR.
- */
-#define OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV_STR	"Heartbeat"
-#define OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG_STR	"AbortedResize"
-#define OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT_STR		"Local"
-#define OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC_STR		"SparseAlloc"
-#define OCFS2_FEATURE_INCOMPAT_EXTENDED_SLOT_MAP_STR	"ExtendedSlotMap"
-#define OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG_STR	"TunefsAbort"
-#define OCFS2_FEATURE_COMPAT_BACKUP_SB_STR		"BackupSuper"
-#define OCFS2_FEATURE_RO_COMPAT_UNWRITTEN_STR		"UnwrittenExtents"
-#define OCFS2_TUNEFS_INPROG_REMOVE_SLOT_STR		"RemoveSlot"
 
 /* define OCFS2_SB for ocfs2-tools */
 #define OCFS2_SB(sb)	(sb)
@@ -109,6 +96,7 @@
 #define OCFS2_DIRENT_FLAG_INCLUDE_EMPTY		0x01
 #define OCFS2_DIRENT_FLAG_INCLUDE_REMOVED	0x02
 #define OCFS2_DIRENT_FLAG_EXCLUDE_DOTS		0x04
+#define OCFS2_DIRENT_FLAG_INCLUDE_TRAILER	0x08
 
 /* Return flags for the chain iterator functions */
 #define OCFS2_CHAIN_CHANGED	0x01
@@ -263,8 +251,8 @@ errcode_t ocfs2_flush(ocfs2_filesys *fs);
 errcode_t ocfs2_close(ocfs2_filesys *fs);
 void ocfs2_freefs(ocfs2_filesys *fs);
 
-void ocfs2_swap_inode_from_cpu(struct ocfs2_dinode *di);
-void ocfs2_swap_inode_to_cpu(struct ocfs2_dinode *di);
+void ocfs2_swap_inode_from_cpu(struct ocfs2_dinode *di, size_t blocksize);
+void ocfs2_swap_inode_to_cpu(struct ocfs2_dinode *di, size_t blocksize);
 errcode_t ocfs2_read_inode(ocfs2_filesys *fs, uint64_t blkno,
 			   char *inode_buf);
 errcode_t ocfs2_write_inode(ocfs2_filesys *fs, uint64_t blkno,
@@ -301,8 +289,17 @@ errcode_t ocfs2_read_journal_superblock(ocfs2_filesys *fs, uint64_t blkno,
 errcode_t ocfs2_write_journal_superblock(ocfs2_filesys *fs, uint64_t blkno,
 					 char *jsb_buf);
 errcode_t ocfs2_make_journal(ocfs2_filesys *fs, uint64_t blkno,
-			     uint32_t clusters);
+			     uint32_t clusters, ocfs2_fs_options *features);
+errcode_t ocfs2_journal_clear_features(journal_superblock_t *jsb,
+				       ocfs2_fs_options *features);
+errcode_t ocfs2_journal_set_features(journal_superblock_t *jsb,
+				     ocfs2_fs_options *features);
+extern size_t ocfs2_journal_tag_bytes(journal_superblock_t *jsb);
+extern uint64_t ocfs2_journal_tag_block(journal_block_tag_t *tag,
+					size_t tag_bytes);
 
+void ocfs2_swap_extent_block_to_cpu(struct ocfs2_extent_block *eb);
+void ocfs2_swap_extent_block_from_cpu(struct ocfs2_extent_block *eb);
 errcode_t ocfs2_read_extent_block(ocfs2_filesys *fs, uint64_t blkno,
        				  char *eb_buf);
 errcode_t ocfs2_read_extent_block_nocheck(ocfs2_filesys *fs, uint64_t blkno,
@@ -311,10 +308,20 @@ errcode_t ocfs2_write_extent_block(ocfs2_filesys *fs, uint64_t blkno,
        				   char *eb_buf);
 errcode_t ocfs2_swap_dir_entries_from_cpu(void *buf, uint64_t bytes);
 errcode_t ocfs2_swap_dir_entries_to_cpu(void *buf, uint64_t bytes);
-errcode_t ocfs2_read_dir_block(ocfs2_filesys *fs, uint64_t block,
-			       void *buf);
-errcode_t ocfs2_write_dir_block(ocfs2_filesys *fs, uint64_t block,
-				void *buf);
+void ocfs2_swap_dir_trailer(struct ocfs2_dir_block_trailer *trailer);
+errcode_t ocfs2_read_dir_block(ocfs2_filesys *fs, struct ocfs2_dinode *di,
+			       uint64_t block, void *buf);
+errcode_t ocfs2_write_dir_block(ocfs2_filesys *fs, struct ocfs2_dinode *di,
+				uint64_t block, void *buf);
+unsigned int ocfs2_dir_trailer_blk_off(ocfs2_filesys *fs);
+struct ocfs2_dir_block_trailer *ocfs2_dir_trailer_from_block(ocfs2_filesys *fs,
+							     void *data);
+int ocfs2_supports_dir_trailer(ocfs2_filesys *fs);
+int ocfs2_dir_has_trailer(ocfs2_filesys *fs, struct ocfs2_dinode *di);
+int ocfs2_skip_dir_trailer(ocfs2_filesys *fs, struct ocfs2_dinode *di,
+			   struct ocfs2_dir_entry *de, unsigned long offset);
+void ocfs2_init_dir_trailer(ocfs2_filesys *fs, struct ocfs2_dinode *di,
+			    uint64_t blkno, void *buf);
 
 errcode_t ocfs2_dir_iterate2(ocfs2_filesys *fs,
 			     uint64_t dir,
@@ -459,9 +466,12 @@ errcode_t ocfs2_chain_force_val(ocfs2_filesys *fs,
 errcode_t ocfs2_chain_add_group(ocfs2_filesys *fs,
 				ocfs2_cached_inode *cinode);
 
+errcode_t ocfs2_init_dir(ocfs2_filesys *fs,
+			 uint64_t dir,
+			 uint64_t parent_dir);
+
 errcode_t ocfs2_expand_dir(ocfs2_filesys *fs,
-			   uint64_t dir,
-			   uint64_t parent_dir);
+			   uint64_t dir);
 
 errcode_t ocfs2_test_inode_allocated(ocfs2_filesys *fs, uint64_t blkno,
 				     int *is_allocated);
@@ -477,7 +487,13 @@ errcode_t ocfs2_new_dir_block(ocfs2_filesys *fs, uint64_t dir_ino,
 errcode_t ocfs2_insert_extent(ocfs2_filesys *fs, uint64_t ino, uint32_t cpos,
 			      uint64_t c_blkno, uint32_t clusters,
 			      uint16_t flag);
+errcode_t ocfs2_cached_inode_insert_extent(ocfs2_cached_inode *ci,
+					   uint32_t cpos, uint64_t c_blkno,
+					   uint32_t clusters, uint16_t flag);
 
+void ocfs2_dinode_new_extent_list(ocfs2_filesys *fs, struct ocfs2_dinode *di);
+void ocfs2_set_inode_data_inline(ocfs2_filesys *fs, struct ocfs2_dinode *di);
+errcode_t ocfs2_convert_inline_data_to_extents(ocfs2_cached_inode *ci);
 errcode_t ocfs2_new_inode(ocfs2_filesys *fs, uint64_t *ino, int mode);
 errcode_t ocfs2_new_system_inode(ocfs2_filesys *fs, uint64_t *ino, int mode, int flags);
 errcode_t ocfs2_delete_inode(ocfs2_filesys *fs, uint64_t ino);
@@ -508,9 +524,17 @@ errcode_t ocfs2_new_clusters(ocfs2_filesys *fs,
 			     uint32_t requested,
 			     uint64_t *start_blkno,
 			     uint32_t *clusters_found);
+errcode_t ocfs2_test_cluster_allocated(ocfs2_filesys *fs, uint32_t cpos,
+				       int *is_allocated);
+errcode_t ocfs2_new_specific_cluster(ocfs2_filesys *fs, uint32_t cpos);
 errcode_t ocfs2_free_clusters(ocfs2_filesys *fs,
 			      uint32_t len,
 			      uint64_t start_blkno);
+errcode_t ocfs2_test_clusters(ocfs2_filesys *fs,
+			      uint32_t len,
+			      uint64_t start_blkno,
+			      int test,
+			      int *matches);
 
 errcode_t ocfs2_lookup(ocfs2_filesys *fs, uint64_t dir, const char *name,
 		       int namelen, char *buf, uint64_t *inode);
@@ -571,6 +595,17 @@ errcode_t ocfs2_write_slot_map_extended(ocfs2_filesys *fs,
 					int num_slots,
 					struct ocfs2_slot_map_extended *se);
 
+/* High level functions for metadata ecc */
+void ocfs2_compute_meta_ecc(ocfs2_filesys *fs, void *data,
+			    struct ocfs2_block_check *bc);
+errcode_t ocfs2_validate_meta_ecc(ocfs2_filesys *fs, void *data,
+				  struct ocfs2_block_check *bc);
+/* Low level checksum compute functions.  Use the high-level ones. */
+extern void ocfs2_block_check_compute(void *data, size_t blocksize,
+				      struct ocfs2_block_check *bc);
+extern errcode_t ocfs2_block_check_validate(void *data, size_t blocksize,
+					    struct ocfs2_block_check *bc);
+
 /* High level */
 errcode_t ocfs2_format_slot_map(ocfs2_filesys *fs);
 errcode_t ocfs2_load_slot_map(ocfs2_filesys *fs,
@@ -583,10 +618,14 @@ enum ocfs2_lock_type ocfs2_get_lock_type(char c);
 char *ocfs2_get_lock_type_string(enum ocfs2_lock_type type);
 
 errcode_t ocfs2_encode_lockres(enum ocfs2_lock_type type, uint64_t blkno,
-			       uint32_t generation, char *lockres);
+			       uint32_t generation, uint64_t parent,
+			       char *lockres);
 
-errcode_t ocfs2_decode_lockres(char *lockres, int len, enum ocfs2_lock_type *type,
-			       uint64_t *blkno, uint32_t *generation);
+errcode_t ocfs2_decode_lockres(char *lockres, enum ocfs2_lock_type *type,
+			       uint64_t *blkno, uint32_t *generation,
+			       uint64_t *parent);
+
+errcode_t ocfs2_printable_lockres(char *lockres, char *name, int len);
 
 /* write the superblock at the specific block. */
 errcode_t ocfs2_write_backup_super(ocfs2_filesys *fs, uint64_t blkno);
@@ -632,6 +671,9 @@ enum ocfs2_feature_levels {
 	OCFS2_FEATURE_LEVEL_MAX_FEATURES,
 };
 
+errcode_t ocfs2_snprint_feature_flags(char *str, size_t size,
+				      ocfs2_fs_options *flags);
+errcode_t ocfs2_snprint_tunefs_flags(char *str, size_t size, uint16_t flags);
 errcode_t ocfs2_parse_feature(const char *opts,
 			      ocfs2_fs_options *feature_flags,
 			      ocfs2_fs_options *reverse_flags);
@@ -644,6 +686,21 @@ errcode_t ocfs2_merge_feature_flags_with_level(ocfs2_fs_options *dest,
 					       ocfs2_fs_options *feature_set,
 					       ocfs2_fs_options *reverse_set);
 
+/*
+ * Get a callback with each feature in feature_set in order.  This will
+ * calculate the dependencies of each feature in feature_set, then call func
+ * once per feature, with only that feature passed to func.
+ */
+void ocfs2_feature_foreach(ocfs2_fs_options *feature_set,
+			   int (*func)(ocfs2_fs_options *feature,
+				       void *user_data),
+			   void *user_data);
+/* The reverse function.  It will call the features in reverse order. */
+void ocfs2_feature_reverse_foreach(ocfs2_fs_options *reverse_set,
+				   int (*func)(ocfs2_fs_options *feature,
+					       void *user_data),
+				   void *user_data);
+
 
 /* These are deprecated names - don't use them */
 int ocfs2_get_backup_super_offset(ocfs2_filesys *fs,
@@ -662,6 +719,11 @@ errcode_t ocfs2_set_backup_super(ocfs2_filesys *fs,
  *
  * ${foo}_in_${bar} is a ceiling function.  clusters_in_blocks will give
  * the number of clusters needed to hold a given number of blocks.
+ *
+ * These functions return UINTxx_MAX when they overflow, but UINTxx_MAX
+ * cannot be used to check overflow.  UINTxx_MAX is a valid value in much
+ * of ocfs2.  The caller is responsible for preventing overflow before
+ * using these functions.
  */
 
 static inline uint64_t ocfs2_clusters_to_blocks(ocfs2_filesys *fs,
@@ -680,18 +742,54 @@ static inline uint32_t ocfs2_blocks_to_clusters(ocfs2_filesys *fs,
 	int b_to_c_bits =
 		OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits -
 		OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits;
+	uint64_t ret = blocks >> b_to_c_bits;
 
-	return (uint32_t)(blocks >> b_to_c_bits);
+	if (ret > UINT32_MAX)
+		ret = UINT32_MAX;
+
+	return (uint32_t)ret;
 }
 
-static inline uint64_t ocfs2_blocks_in_bytes(ocfs2_filesys *fs, uint64_t bytes)
+static inline uint64_t ocfs2_clusters_to_bytes(ocfs2_filesys *fs,
+					       uint32_t clusters)
 {
-	uint64_t ret = bytes + fs->fs_blocksize - 1;
+	uint64_t ret = clusters;
 
-	if (ret < bytes) /* deal with wrapping */
-		return UINT64_MAX;
+	ret = ret << OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits;
+	if (ret < clusters)
+		ret = UINT64_MAX;
 
-	return ret >> OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits;
+	return ret;
+}
+
+static inline uint32_t ocfs2_bytes_to_clusters(ocfs2_filesys *fs,
+					       uint64_t bytes)
+{
+	uint64_t ret =
+		bytes >> OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits;
+
+	if (ret > UINT32_MAX)
+		ret = UINT32_MAX;
+
+	return (uint32_t)ret;
+}
+
+static inline uint64_t ocfs2_blocks_to_bytes(ocfs2_filesys *fs,
+					     uint64_t blocks)
+{
+	uint64_t ret =
+		blocks << OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits;
+
+	if (ret < blocks)
+		ret = UINT64_MAX;
+
+	return ret;
+}
+
+static inline uint64_t ocfs2_bytes_to_blocks(ocfs2_filesys *fs,
+					     uint64_t bytes)
+{
+	return bytes >> OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits;
 }
 
 static inline uint32_t ocfs2_clusters_in_blocks(ocfs2_filesys *fs, 
@@ -704,9 +802,40 @@ static inline uint32_t ocfs2_clusters_in_blocks(ocfs2_filesys *fs,
 	if (ret < blocks) /* deal with wrapping */
 		ret = UINT64_MAX;
 
-	return (uint32_t)(ret >> c_to_b_bits);
+	ret = ret >> c_to_b_bits;
+	if (ret > UINT32_MAX)
+		ret = UINT32_MAX;
+
+	return (uint32_t)ret;
 }
 
+static inline uint32_t ocfs2_clusters_in_bytes(ocfs2_filesys *fs,
+					       uint64_t bytes)
+{
+	uint64_t ret = bytes + fs->fs_clustersize - 1;
+
+	if (ret < bytes) /* deal with wrapping */
+		ret = UINT64_MAX;
+
+	ret = ret >> OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits;
+	if (ret > UINT32_MAX)
+		ret = UINT32_MAX;
+
+	return (uint32_t)ret;
+}
+
+static inline uint64_t ocfs2_blocks_in_bytes(ocfs2_filesys *fs,
+					     uint64_t bytes)
+{
+	uint64_t ret = bytes + fs->fs_blocksize - 1;
+
+	if (ret < bytes) /* deal with wrapping */
+		return UINT64_MAX;
+
+	return ret >> OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits;
+}
+
+
 /* given a cluster offset, calculate which block group it belongs to
  * and return that block offset. */
 static inline uint64_t ocfs2_which_cluster_group(ocfs2_filesys *fs,
@@ -822,6 +951,27 @@ static inline int ocfs2_uses_extended_slot_map(struct ocfs2_super_block *osb)
 	return 0;
 }
 
+static inline int ocfs2_support_inline_data(struct ocfs2_super_block *osb)
+{
+	if (osb->s_feature_incompat & OCFS2_FEATURE_INCOMPAT_INLINE_DATA)
+		return 1;
+	return 0;
+}
+
+static inline int ocfs2_meta_ecc(struct ocfs2_super_block *osb)
+{
+	if (OCFS2_HAS_INCOMPAT_FEATURE(osb,
+				       OCFS2_FEATURE_INCOMPAT_META_ECC))
+		return 1;
+	return 0;
+}
+
+static inline int ocfs2_support_xattr(struct ocfs2_super_block *osb)
+{
+	if (osb->s_feature_incompat & OCFS2_FEATURE_INCOMPAT_XATTR)
+		return 1;
+	return 0;
+}
 
 /*
  * shamelessly lifted from the kernel
@@ -941,4 +1091,55 @@ errcode_t ocfs2_block_iterate_inode(ocfs2_filesys *fs,
 						void *priv_data),
 				    void *priv_data);
 
+uint32_t ocfs2_xattr_uuid_hash(unsigned char *uuid);
+uint32_t ocfs2_xattr_name_hash(uint32_t uuid_hash, const char *name,
+			       int name_len);
+int ocfs2_xattr_find_leaf(ocfs2_filesys *fs, struct ocfs2_xattr_block *xb,
+			  uint32_t cpos, char **leaf_buf);
+uint16_t ocfs2_xattr_buckets_per_cluster(ocfs2_filesys *fs);
+uint16_t ocfs2_blocks_per_xattr_bucket(ocfs2_filesys *fs);
+void ocfs2_swap_xattrs_to_cpu(struct ocfs2_xattr_header *xh);
+void ocfs2_swap_xattrs_from_cpu(struct ocfs2_xattr_header *xh);
+void ocfs2_swap_xattr_block_to_cpu(struct ocfs2_xattr_block *xb);
+void ocfs2_swap_xattr_block_from_cpu(struct ocfs2_xattr_block *xb);
+errcode_t ocfs2_read_xattr_block(ocfs2_filesys *fs,
+				 uint64_t blkno,
+				 char *xb_buf);
+errcode_t ocfs2_write_xattr_block(ocfs2_filesys *fs,
+				  uint64_t blkno,
+				  char *xb_buf);
+errcode_t ocfs2_xattr_get_rec(ocfs2_filesys *fs,
+			      struct ocfs2_xattr_block *xb,
+			      uint32_t name_hash,
+			      uint64_t *p_blkno,
+			      uint32_t *e_cpos,
+			      uint32_t *num_clusters);
+uint16_t ocfs2_xattr_value_real_size(uint16_t name_len, uint16_t value_len);
+uint16_t ocfs2_xattr_min_offset(struct ocfs2_xattr_header *xh, uint16_t size);
+uint16_t ocfs2_xattr_name_value_len(struct ocfs2_xattr_header *xh);
+errcode_t ocfs2_read_xattr_bucket(ocfs2_filesys *fs,
+				  uint64_t blkno,
+				  char *bucket_buf);
+errcode_t ocfs2_write_xattr_bucket(ocfs2_filesys *fs,
+				   uint64_t blkno,
+				   char *bucket_buf);
+errcode_t ocfs2_xattr_value_truncate(ocfs2_filesys *fs,
+				     struct ocfs2_xattr_value_root *xv);
+errcode_t ocfs2_xattr_tree_truncate(ocfs2_filesys *fs,
+				    struct ocfs2_xattr_tree_root *xt);
+errcode_t ocfs2_extent_iterate_xattr(ocfs2_filesys *fs,
+				     struct ocfs2_extent_list *el,
+				     uint64_t last_eb_blk,
+				     int flags,
+				     int (*func)(ocfs2_filesys *fs,
+						struct ocfs2_extent_rec *rec,
+						int tree_depth,
+						uint32_t ccount,
+						uint64_t ref_blkno,
+						int ref_recno,
+						void *priv_data),
+				     void *priv_data,
+				     int *changed);
+errcode_t ocfs2_delete_xattr_block(ocfs2_filesys *fs, uint64_t blkno);
+
 #endif  /* _FILESYS_H */
