@@ -85,7 +85,7 @@ static void ocfs2_swap_extent_block_header(struct ocfs2_extent_block *eb)
 	eb->h_next_leaf_blk = bswap_64(eb->h_next_leaf_blk);
 }
 
-static void ocfs2_swap_extent_block_from_cpu(struct ocfs2_extent_block *eb)
+void ocfs2_swap_extent_block_from_cpu(struct ocfs2_extent_block *eb)
 {
 	if (cpu_is_little_endian)
 		return;
@@ -93,7 +93,8 @@ static void ocfs2_swap_extent_block_from_cpu(struct ocfs2_extent_block *eb)
 	ocfs2_swap_extent_block_header(eb);
 	ocfs2_swap_extent_list_from_cpu(&eb->h_list);
 }
-static void ocfs2_swap_extent_block_to_cpu(struct ocfs2_extent_block *eb)
+
+void ocfs2_swap_extent_block_to_cpu(struct ocfs2_extent_block *eb)
 {
 	if (cpu_is_little_endian)
 		return;
@@ -124,6 +125,10 @@ errcode_t ocfs2_read_extent_block_nocheck(ocfs2_filesys *fs,
 
 	eb = (struct ocfs2_extent_block *)blk;
 
+	ret = ocfs2_validate_meta_ecc(fs, blk, &eb->h_check);
+	if (ret)
+		goto out;
+
 	if (memcmp(eb->h_signature, OCFS2_EXTENT_BLOCK_SIGNATURE,
 		   strlen(OCFS2_EXTENT_BLOCK_SIGNATURE))) {
 		ret = OCFS2_ET_BAD_EXTENT_BLOCK_MAGIC;
@@ -179,6 +184,7 @@ errcode_t ocfs2_write_extent_block(ocfs2_filesys *fs, uint64_t blkno,
 	eb = (struct ocfs2_extent_block *) blk;
 	ocfs2_swap_extent_block_from_cpu(eb);
 
+	ocfs2_compute_meta_ecc(fs, blk, &eb->h_check);
 	ret = io_write_block(fs->fs_io, blkno, 1, blk);
 	if (ret)
 		goto out;
@@ -372,6 +378,77 @@ out:
 	return iret;
 }
 
+errcode_t ocfs2_extent_iterate_xattr(ocfs2_filesys *fs,
+				     struct ocfs2_extent_list *el,
+				     uint64_t last_eb_blk,
+				     int flags,
+				     int (*func)(ocfs2_filesys *fs,
+						struct ocfs2_extent_rec *rec,
+						int tree_depth,
+						uint32_t ccount,
+						uint64_t ref_blkno,
+						int ref_recno,
+						void *priv_data),
+				     void *priv_data,
+				     int *changed)
+{
+	int i;
+	int iret = 0;
+	errcode_t ret;
+	struct extent_context ctxt;
+
+	if (el->l_tree_depth) {
+		ret = ocfs2_malloc0(sizeof(char *) * el->l_tree_depth,
+				    &ctxt.eb_bufs);
+		if (ret)
+			goto out;
+
+		ret = ocfs2_malloc0(fs->fs_blocksize *
+				    el->l_tree_depth,
+				    &ctxt.eb_bufs[0]);
+		if (ret)
+			goto out_eb_bufs;
+
+		for (i = 1; i < el->l_tree_depth; i++) {
+			ctxt.eb_bufs[i] = ctxt.eb_bufs[0] +
+				i * fs->fs_blocksize;
+		}
+	} else
+		ctxt.eb_bufs = NULL;
+
+	ctxt.fs = fs;
+	ctxt.func = func;
+	ctxt.priv_data = priv_data;
+	ctxt.flags = flags;
+	ctxt.ccount = 0;
+	ctxt.last_eb_blkno = 0;
+	ctxt.last_eb_cpos = 0;
+
+	ret = 0;
+	iret |= extent_iterate_el(el, 0, &ctxt);
+	if (iret & OCFS2_EXTENT_ERROR)
+		ret = ctxt.errcode;
+
+	if (iret & OCFS2_EXTENT_ABORT)
+		goto out_abort;
+
+	if (last_eb_blk != ctxt.last_eb_blkno) {
+		last_eb_blk = ctxt.last_eb_blkno;
+		iret |= OCFS2_EXTENT_CHANGED;
+	}
+
+out_abort:
+	if (!ret && (iret & OCFS2_EXTENT_CHANGED))
+		*changed = 1;
+out_eb_bufs:
+	if (ctxt.eb_bufs) {
+		if (ctxt.eb_bufs[0])
+			ocfs2_free(&ctxt.eb_bufs[0]);
+		ocfs2_free(&ctxt.eb_bufs);
+	}
+out:
+	return ret;
+}
 
 errcode_t ocfs2_extent_iterate_inode(ocfs2_filesys *fs,
 				     struct ocfs2_dinode *inode,
