@@ -127,7 +127,10 @@ static void ocfs2_swap_inode_second(struct ocfs2_dinode *di)
 		sb->s_blocksize_bits      = bswap_32(sb->s_blocksize_bits);
 		sb->s_clustersize_bits    = bswap_32(sb->s_clustersize_bits);
 		sb->s_max_slots           = bswap_16(sb->s_max_slots);
+		sb->s_tunefs_flag         = bswap_16(sb->s_tunefs_flag);
+		sb->s_uuid_hash           = bswap_32(sb->s_uuid_hash);
 		sb->s_first_cluster_group = bswap_64(sb->s_first_cluster_group);
+		sb->s_xattr_inline_size   = bswap_16(sb->s_xattr_inline_size);
 
 	} else if (di->i_flags & OCFS2_LOCAL_ALLOC_FL) {
 		struct ocfs2_local_alloc *la = &di->id2.i_lab;
@@ -148,6 +151,10 @@ static void ocfs2_swap_inode_second(struct ocfs2_dinode *di)
 
 		tl->tl_count = bswap_16(tl->tl_count);
 		tl->tl_used  = bswap_16(tl->tl_used);
+	} else if (di->i_dyn_features & OCFS2_INLINE_DATA_FL) {
+		struct ocfs2_inline_data *id = &di->id2.i_data;
+
+		id->id_count = bswap_16(id->id_count);
 	}
 }
 
@@ -156,6 +163,7 @@ static void ocfs2_swap_inode_first(struct ocfs2_dinode *di)
 	di->i_generation    = bswap_32(di->i_generation);
 	di->i_suballoc_slot = bswap_16(di->i_suballoc_slot);
 	di->i_suballoc_bit  = bswap_16(di->i_suballoc_bit);
+	di->i_xattr_inline_size = bswap_16(di->i_xattr_inline_size);
 	di->i_clusters      = bswap_32(di->i_clusters);
 	di->i_uid           = bswap_32(di->i_uid);
 	di->i_gid           = bswap_32(di->i_gid);
@@ -174,6 +182,9 @@ static void ocfs2_swap_inode_first(struct ocfs2_dinode *di)
 	di->i_ctime_nsec    = bswap_32(di->i_ctime_nsec);
 	di->i_mtime_nsec    = bswap_32(di->i_mtime_nsec);
 	di->i_attr          = bswap_32(di->i_attr);
+	di->i_orphaned_slot = bswap_16(di->i_orphaned_slot);
+	di->i_dyn_features  = bswap_16(di->i_dyn_features);
+	di->i_xattr_loc     = bswap_64(di->i_xattr_loc);
 }
 
 static int has_extents(struct ocfs2_dinode *di)
@@ -185,23 +196,44 @@ static int has_extents(struct ocfs2_dinode *di)
 	/* i_flags doesn't indicate when id2 is a fast symlink */
 	if (S_ISLNK(di->i_mode) && di->i_size && di->i_clusters == 0)
 		return 0;
+	if (di->i_dyn_features & OCFS2_INLINE_DATA_FL)
+		return 0;
 
 	return 1;
 }
 
-void ocfs2_swap_inode_from_cpu(struct ocfs2_dinode *di)
+static inline void ocfs2_swap_inline_dir(struct ocfs2_dinode *di,
+					 int to_cpu)
+{
+	void *de_buf = di->id2.i_data.id_data;
+	uint64_t bytes = di->id2.i_data.id_count;
+
+	if (to_cpu)
+		ocfs2_swap_dir_entries_to_cpu(de_buf, bytes);
+	else
+		ocfs2_swap_dir_entries_from_cpu(de_buf, bytes);
+}
+
+void ocfs2_swap_inode_from_cpu(struct ocfs2_dinode *di, size_t blocksize)
 {
 	if (cpu_is_little_endian)
 		return;
 
+	if (di->i_dyn_features & OCFS2_INLINE_XATTR_FL) {
+		struct ocfs2_xattr_header *xh = (struct ocfs2_xattr_header *)
+			((void *)di + blocksize - di->i_xattr_inline_size);
+		ocfs2_swap_xattrs_from_cpu(xh);
+	}
 	if (has_extents(di))
 		ocfs2_swap_extent_list_from_cpu(&di->id2.i_list);
+	if (di->i_dyn_features & OCFS2_INLINE_DATA_FL && S_ISDIR(di->i_mode))
+		ocfs2_swap_inline_dir(di, 0);
 	ocfs2_swap_inode_third(di);
 	ocfs2_swap_inode_second(di);
 	ocfs2_swap_inode_first(di);
 }
 
-void ocfs2_swap_inode_to_cpu(struct ocfs2_dinode *di)
+void ocfs2_swap_inode_to_cpu(struct ocfs2_dinode *di, size_t blocksize)
 {
 	if (cpu_is_little_endian)
 		return;
@@ -209,8 +241,15 @@ void ocfs2_swap_inode_to_cpu(struct ocfs2_dinode *di)
 	ocfs2_swap_inode_first(di);
 	ocfs2_swap_inode_second(di);
 	ocfs2_swap_inode_third(di);
+	if (di->i_dyn_features & OCFS2_INLINE_DATA_FL && S_ISDIR(di->i_mode))
+		ocfs2_swap_inline_dir(di, 1);
 	if (has_extents(di))
 		ocfs2_swap_extent_list_to_cpu(&di->id2.i_list);
+	if (di->i_dyn_features & OCFS2_INLINE_XATTR_FL) {
+		struct ocfs2_xattr_header *xh = (struct ocfs2_xattr_header *)
+			((void *)di + blocksize - di->i_xattr_inline_size);
+		ocfs2_swap_xattrs_to_cpu(xh);
+	}
 }
 
 errcode_t ocfs2_read_inode(ocfs2_filesys *fs, uint64_t blkno,
@@ -233,6 +272,9 @@ errcode_t ocfs2_read_inode(ocfs2_filesys *fs, uint64_t blkno,
 		goto out;
 
 	di = (struct ocfs2_dinode *)blk;
+	ret = ocfs2_validate_meta_ecc(fs, blk, &di->i_check);
+	if (ret)
+		goto out;
 
 	ret = OCFS2_ET_BAD_INODE_MAGIC;
 	if (memcmp(di->i_signature, OCFS2_INODE_SIGNATURE,
@@ -242,7 +284,7 @@ errcode_t ocfs2_read_inode(ocfs2_filesys *fs, uint64_t blkno,
 	memcpy(inode_buf, blk, fs->fs_blocksize);
 
 	di = (struct ocfs2_dinode *) inode_buf;
-	ocfs2_swap_inode_to_cpu(di);
+	ocfs2_swap_inode_to_cpu(di, fs->fs_blocksize);
 
 	ret = 0;
 out:
@@ -272,8 +314,9 @@ errcode_t ocfs2_write_inode(ocfs2_filesys *fs, uint64_t blkno,
 	memcpy(blk, inode_buf, fs->fs_blocksize);
 
 	di = (struct ocfs2_dinode *)blk;
-	ocfs2_swap_inode_from_cpu(di);
+	ocfs2_swap_inode_from_cpu(di, fs->fs_blocksize);
 
+	ocfs2_compute_meta_ecc(fs, blk, &di->i_check);
 	ret = io_write_block(fs->fs_io, blkno, 1, blk);
 	if (ret)
 		goto out;
