@@ -34,6 +34,74 @@
 #include "ocfs2/byteorder.h"
 #include "ocfs2/ocfs2.h"
 
+size_t ocfs2_journal_tag_bytes(journal_superblock_t *jsb)
+{
+	if (JBD2_HAS_INCOMPAT_FEATURE(jsb, JBD2_FEATURE_INCOMPAT_64BIT))
+		return JBD2_TAG_SIZE64;
+	else
+		return JBD2_TAG_SIZE32;
+}
+
+uint64_t ocfs2_journal_tag_block(journal_block_tag_t *tag, size_t tag_bytes)
+{
+	uint64_t blockno = be32_to_cpu(tag->t_blocknr);
+	if (tag_bytes > JBD2_TAG_SIZE32)
+		blockno |= (uint64_t)be32_to_cpu(tag->t_blocknr_high) << 32;
+	return blockno;
+}
+
+/* Returns true if we support these journal features */
+static int ocfs2_journal_check_available_features(journal_superblock_t *jsb,
+						  ocfs2_fs_options *features)
+{
+	/* ocfs2 has never supported a superblock other than _V2 */
+	if (jsb->s_header.h_blocktype != JBD2_SUPERBLOCK_V2)
+		return 0;
+
+	if ((features->opt_compat & JBD2_KNOWN_COMPAT_FEATURES) !=
+	    features->opt_compat)
+		return 0;
+
+	if ((features->opt_ro_compat & JBD2_KNOWN_ROCOMPAT_FEATURES) !=
+	    features->opt_ro_compat)
+		return 0;
+
+	if ((features->opt_incompat & JBD2_KNOWN_INCOMPAT_FEATURES) !=
+	    features->opt_incompat)
+		return 0;
+
+	return 1;
+}
+
+errcode_t ocfs2_journal_set_features(journal_superblock_t *jsb,
+				     ocfs2_fs_options *features)
+{
+
+	if (!ocfs2_journal_check_available_features(jsb, features))
+		return OCFS2_ET_UNSUPP_FEATURE;
+
+
+	jsb->s_feature_compat |= features->opt_compat;
+	jsb->s_feature_ro_compat |= features->opt_ro_compat;
+	jsb->s_feature_incompat |= features->opt_incompat;
+
+	return 0;
+}
+
+errcode_t ocfs2_journal_clear_features(journal_superblock_t *jsb,
+				       ocfs2_fs_options *features)
+{
+
+	if (!ocfs2_journal_check_available_features(jsb, features))
+		return OCFS2_ET_UNSUPP_FEATURE;
+
+	jsb->s_feature_compat &= ~(features->opt_compat);
+	jsb->s_feature_ro_compat &= ~(features->opt_ro_compat);
+	jsb->s_feature_incompat &= ~(features->opt_incompat);
+
+	return 0;
+}
+
 void ocfs2_swap_journal_superblock(journal_superblock_t *jsb)
 {
 	if (cpu_is_big_endian)
@@ -70,8 +138,8 @@ errcode_t ocfs2_init_journal_superblock(ocfs2_filesys *fs, char *buf,
 		return OCFS2_ET_JOURNAL_TOO_SMALL;
 
 	memset(buf, 0, buflen);
-	jsb->s_header.h_magic     = JFS_MAGIC_NUMBER;
-	jsb->s_header.h_blocktype = JFS_SUPERBLOCK_V2;
+	jsb->s_header.h_magic     = JBD2_MAGIC_NUMBER;
+	jsb->s_header.h_blocktype = JBD2_SUPERBLOCK_V2;
 
 	jsb->s_blocksize = fs->fs_blocksize;
 	jsb->s_maxlen    = jrnl_size_in_blks;
@@ -98,10 +166,12 @@ errcode_t ocfs2_init_journal_superblock(ocfs2_filesys *fs, char *buf,
  */
 static errcode_t ocfs2_create_journal_superblock(ocfs2_filesys *fs,
 						 uint32_t size,
+						 ocfs2_fs_options *features,
 						 char  **ret_jsb)
 {
 	errcode_t retval;
 	char *buf = NULL;
+	journal_superblock_t *jsb;
 
 	*ret_jsb = NULL;
 
@@ -112,6 +182,11 @@ static errcode_t ocfs2_create_journal_superblock(ocfs2_filesys *fs,
 	if (retval)
 		goto bail;
 
+	jsb = (journal_superblock_t *)buf;
+	retval = ocfs2_journal_set_features(jsb, features);
+	if (retval)
+		goto bail;
+
 #if 0 /* Someday */
 	/*
 	 * If we're creating an external journal device, we need to
@@ -157,7 +232,7 @@ errcode_t ocfs2_read_journal_superblock(ocfs2_filesys *fs, uint64_t blkno,
 	disk = (journal_superblock_t *)blk;
 	jsb = (journal_superblock_t *)jsb_buf;
 
-	if (disk->s_header.h_magic != htonl(JFS_MAGIC_NUMBER)) {
+	if (disk->s_header.h_magic != htonl(JBD2_MAGIC_NUMBER)) {
 		ret = OCFS2_ET_BAD_JOURNAL_SUPERBLOCK_MAGIC;
 		goto out;
 	}
@@ -165,6 +240,16 @@ errcode_t ocfs2_read_journal_superblock(ocfs2_filesys *fs, uint64_t blkno,
 	memcpy(jsb_buf, blk, fs->fs_blocksize);
 	ocfs2_swap_journal_superblock(jsb);
 
+	if (JBD2_HAS_INCOMPAT_FEATURE(jsb, ~JBD2_KNOWN_INCOMPAT_FEATURES)) {
+		ret = OCFS2_ET_UNSUPP_FEATURE;
+		goto out;
+	}
+
+	if (JBD2_HAS_RO_COMPAT_FEATURE(jsb, ~JBD2_KNOWN_ROCOMPAT_FEATURES)) {
+		ret = OCFS2_ET_RO_UNSUPP_FEATURE;
+		goto out;
+	}
+
 	ret = 0;
 out:
 	ocfs2_free(&blk);
@@ -209,7 +294,8 @@ out:
 }
 
 static errcode_t ocfs2_format_journal(ocfs2_filesys *fs,
-				      ocfs2_cached_inode *ci)
+				      ocfs2_cached_inode *ci,
+				      ocfs2_fs_options *features)
 {
 	errcode_t ret = 0;
 	char *buf = NULL, *jsb_buf = NULL;
@@ -234,7 +320,8 @@ static errcode_t ocfs2_format_journal(ocfs2_filesys *fs,
 	}
 
 	jrnl_blocks = ocfs2_clusters_to_blocks(fs, ci->ci_inode->i_clusters);
-	ret = ocfs2_create_journal_superblock(fs, jrnl_blocks, &jsb_buf);
+	ret = ocfs2_create_journal_superblock(fs, jrnl_blocks, features,
+					      &jsb_buf);
 	if (ret)
 		goto out;
 
@@ -254,7 +341,7 @@ out:
 }
 
 errcode_t ocfs2_make_journal(ocfs2_filesys *fs, uint64_t blkno,
-			     uint32_t clusters)
+			     uint32_t clusters, ocfs2_fs_options *features)
 {
 	errcode_t ret = 0;
 	ocfs2_cached_inode *ci = NULL;
@@ -318,7 +405,7 @@ errcode_t ocfs2_make_journal(ocfs2_filesys *fs, uint64_t blkno,
 		}
 	}
 
-	ret = ocfs2_format_journal(fs, ci);
+	ret = ocfs2_format_journal(fs, ci, features);
 out:
 	if (ci)
 		ocfs2_free_cached_inode(fs, ci);
