@@ -102,8 +102,10 @@ out:
 errcode_t ocfs2_read_super(ocfs2_filesys *fs, uint64_t superblock, char *sb)
 {
 	errcode_t ret;
-	char *blk;
-	struct ocfs2_dinode *di;
+	char *blk, *swapblk;
+	struct ocfs2_dinode *di, *orig_super;
+	int orig_blocksize;
+	int blocksize = io_get_blksize(fs->fs_io);
 
 	ret = ocfs2_malloc_block(fs->fs_io, &blk);
 	if (ret)
@@ -112,6 +114,7 @@ errcode_t ocfs2_read_super(ocfs2_filesys *fs, uint64_t superblock, char *sb)
 	ret = ocfs2_read_blocks(fs, superblock, 1, blk);
 	if (ret)
 		goto out_blk;
+
 	di = (struct ocfs2_dinode *)blk;
 
 	ret = OCFS2_ET_BAD_MAGIC;
@@ -119,8 +122,33 @@ errcode_t ocfs2_read_super(ocfs2_filesys *fs, uint64_t superblock, char *sb)
 		   strlen(OCFS2_SUPER_BLOCK_SIGNATURE)))
 		goto out_blk;
 
-	ocfs2_swap_inode_to_cpu(di);
+	/*
+	 * We want to use the latest superblock to validate.  We need
+	 * a local-endian copy in fs->fs_super, and the unswapped copy to
+	 * check in blk.  ocfs2_validate_meta_ecc() uses fs->fs_super and
+	 * fs->fs_blocksize.
+	 */
+	ret = ocfs2_malloc_block(fs->fs_io, &swapblk);
+	if (ret)
+		goto out_blk;
 
+	memcpy(swapblk, blk, blocksize);
+	orig_super = fs->fs_super;
+	orig_blocksize = fs->fs_blocksize;
+	fs->fs_super = (struct ocfs2_dinode *)swapblk;
+	fs->fs_blocksize = blocksize;
+	ocfs2_swap_inode_to_cpu(fs->fs_super, fs->fs_blocksize);
+
+	ret = ocfs2_validate_meta_ecc(fs, blk, &di->i_check);
+
+	fs->fs_super = orig_super;
+	fs->fs_blocksize = orig_blocksize;
+	ocfs2_free(&swapblk);
+
+	if (ret)
+		goto out_blk;
+
+	ocfs2_swap_inode_to_cpu(di, fs->fs_blocksize);
 	if (!sb)
 		fs->fs_super = di;
 	else {
@@ -288,7 +316,8 @@ errcode_t ocfs2_open(const char *name, int flags,
 				io_set_blksize(fs->fs_io, block_size);
 				ret = ocfs2_read_super(fs, (uint64_t)superblock,
 						       NULL);
-				if (ret == OCFS2_ET_BAD_MAGIC)
+				if ((ret == OCFS2_ET_BAD_MAGIC) ||
+				    (ret == OCFS2_ET_IO))
 					continue;
 				break;
 			}
