@@ -28,7 +28,8 @@
 static State *get_state(int argc, char **argv);
 static int get_number(char *arg, uint64_t *res);
 static void parse_journal_opts(char *progname, const char *opts,
-			       uint64_t *journal_size_in_bytes);
+			       uint64_t *journal_size_in_bytes,
+			       int *journal64);
 static void usage(const char *progname);
 static void version(const char *progname);
 static void fill_defaults(State *s);
@@ -81,10 +82,12 @@ static AllocGroup * initialize_alloc_group(State *s, const char *name,
 					   uint64_t blkno,
 					   uint16_t chain, uint16_t cpg,
 					   uint16_t bpc);
-static void create_lost_found_dir(State *s);
-static void format_journals(State *s);
-static void format_slotmap(State *s);
-static int format_backup_super(State *s);
+static void create_lost_found_dir(State *s, ocfs2_filesys *fs);
+static void format_journals(State *s, ocfs2_filesys *fs);
+static void format_slotmap(State *s, ocfs2_filesys *fs);
+static int format_backup_super(State *s, ocfs2_filesys *fs);
+static void mkfs_compute_meta_ecc(State *s, void *data,
+				  struct ocfs2_block_check *bc);
 
 extern char *optarg;
 extern int optind, opterr, optopt;
@@ -183,6 +186,127 @@ static int hb_dev_skip(State *s, int system_inode)
 	return ret;
 }
 
+static void fill_fake_fs(State *s, ocfs2_filesys *fake_fs, void *buf)
+{
+	memset(buf, 0, s->blocksize);
+	memset(fake_fs, 0, sizeof(ocfs2_filesys));
+
+	fake_fs->fs_super = buf;
+	fake_fs->fs_blocksize = s->blocksize;
+	fake_fs->fs_clustersize = s->cluster_size;
+
+	OCFS2_RAW_SB(fake_fs->fs_super)->s_feature_incompat =
+		s->feature_flags.opt_incompat;
+	OCFS2_RAW_SB(fake_fs->fs_super)->s_feature_ro_compat =
+		s->feature_flags.opt_ro_compat;
+	OCFS2_RAW_SB(fake_fs->fs_super)->s_feature_compat =
+		s->feature_flags.opt_compat;
+}
+
+static void mkfs_init_dir_trailer(State *s, DirData *dir, void *buf)
+{
+	char super_buf[OCFS2_MAX_BLOCKSIZE];
+	ocfs2_filesys fake_fs;
+	struct ocfs2_dir_entry *de;
+	struct ocfs2_dinode fake_di = {
+		.i_blkno = dir->record->fe_off >> s->blocksize_bits,
+	};
+	uint64_t blkno = dir->record->extent_off;
+
+	/* Find out how far we are in our directory */
+	blkno += ((char *)buf) - ((char *)dir->buf);
+	blkno >>= s->blocksize_bits;
+
+	fill_fake_fs(s, &fake_fs, super_buf);
+
+	if (ocfs2_supports_dir_trailer(&fake_fs)) {
+		de = buf;
+		de->rec_len = ocfs2_dir_trailer_blk_off(&fake_fs);
+		ocfs2_init_dir_trailer(&fake_fs, &fake_di, blkno, buf);
+	}
+}
+
+static inline uint32_t system_dir_bytes_needed(State *s)
+{
+	int each = OCFS2_DIR_REC_LEN(SYSTEM_FILE_NAME_MAX);
+
+	return each * sys_blocks_needed(s->initial_slots);
+}
+
+static void finish_normal_format(State *s)
+{
+	errcode_t ret;
+	int num;
+	ocfs2_filesys *fs;
+
+	/* These routines use libocfs2 to do their work. We
+	 * don't share an ocfs2_filesys context between the
+	 * journal format and the lost+found create so that
+	 * the library can use the journal for the latter in
+	 * future revisions. */
+
+	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
+	if (ret) {
+		com_err(s->progname, ret,
+			"while opening file system for final "
+			"operations.");
+		clear_both_ends(s);
+		exit(1);
+	}
+
+	if (!s->no_backup_super) {
+		ret = io_init_cache(fs->fs_io,
+				    ocfs2_extent_recs_per_eb(fs->fs_blocksize));
+		if (ret)
+			com_err(s->progname, ret,
+				"while initializing the I/O cache.  Continuing "
+				"without a cache (safe, but slower)");
+
+		if (!s->quiet)
+			printf("Writing backup superblock: ");
+
+		num = format_backup_super(s, fs);
+		if (!s->quiet)
+			printf("%d block(s)\n", num);
+
+		io_destroy_cache(fs->fs_io);
+	}
+
+	/* io cache is disabled during journal format for performance reasons */
+	if (!s->quiet)
+		printf("Formatting Journals: ");
+
+	format_journals(s, fs);
+
+	if (!s->quiet)
+		printf("done\n");
+
+	ret = io_init_cache(fs->fs_io,
+			    ocfs2_extent_recs_per_eb(fs->fs_blocksize));
+	if (ret)
+		com_err(s->progname, ret,
+			"while initializing the I/O cache.  Continuing "
+			"without a cache (safe, but slower)");
+
+	if (!s->quiet)
+		printf("Formatting slot map: ");
+
+	format_slotmap(s, fs);
+
+	if (!s->quiet)
+		printf("done\n");
+
+	if (!s->quiet)
+		printf("Writing lost+found: ");
+
+	create_lost_found_dir(s, fs);
+
+	if (!s->quiet)
+		printf("done\n");
+
+	ocfs2_close(fs);
+}
+
 int
 main(int argc, char **argv)
 {
@@ -315,9 +439,13 @@ main(int argc, char **argv)
 
 	superblock_rec.fe_off = (uint64_t)OCFS2_SUPER_BLOCK_BLKNO << s->blocksize_bits;
 
-	alloc_from_bitmap (s, 1, s->global_bm,
-			   &root_dir_rec.extent_off,
-			   &root_dir_rec.extent_len);
+	if (!s->inline_data) {
+		alloc_from_bitmap(s, 1, s->global_bm,
+				  &root_dir_rec.extent_off,
+				  &root_dir_rec.extent_len);
+		root_dir_rec.dir_data = NULL;
+	} else
+		root_dir_rec.dir_data = root_dir;
 
 	root_dir_rec.fe_off = alloc_inode(s, &root_dir_rec.suballoc_bit);
 	root_dir->record = &root_dir_rec;
@@ -325,10 +453,16 @@ main(int argc, char **argv)
 	add_entry_to_directory(s, root_dir, ".", root_dir_rec.fe_off, OCFS2_FT_DIR);
 	add_entry_to_directory(s, root_dir, "..", root_dir_rec.fe_off, OCFS2_FT_DIR);
 
-	need = system_dir_blocks_needed(s) << s->blocksize_bits;
-	alloc_bytes_from_bitmap(s, need, s->global_bm,
-				&system_dir_rec.extent_off,
-				&system_dir_rec.extent_len);
+	need = system_dir_bytes_needed(s);
+	if (!s->inline_data || need > ocfs2_max_inline_data(s->blocksize)) {
+		need = system_dir_blocks_needed(s) << s->blocksize_bits;
+		alloc_bytes_from_bitmap(s, need, s->global_bm,
+					&system_dir_rec.extent_off,
+					&system_dir_rec.extent_len);
+		system_dir_rec.dir_data = NULL;
+	} else
+		system_dir_rec.dir_data = system_dir;
+
 	system_dir_rec.fe_off = alloc_inode(s, &system_dir_rec.suballoc_bit);
 	system_dir->record = &system_dir_rec;
 	add_entry_to_directory(s, system_dir, ".", system_dir_rec.fe_off, OCFS2_FT_DIR);
@@ -365,9 +499,14 @@ main(int argc, char **argv)
 		for (i = 0; i < s->initial_slots; ++i) {
 			tmprec = &record[ORPHAN_DIR_SYSTEM_INODE][i];
 			orphan_dir[i]->record = tmprec;
-			alloc_from_bitmap(s, 1, s->global_bm,
-					  &tmprec->extent_off,
-					  &tmprec->extent_len);
+			if (!s->inline_data) {
+				alloc_from_bitmap(s, 1, s->global_bm,
+						  &tmprec->extent_off,
+						  &tmprec->extent_len);
+				tmprec->dir_data = NULL;
+			} else
+				tmprec->dir_data = orphan_dir[i];
+
 			add_entry_to_directory(s, orphan_dir[i], ".",
 					       tmprec->fe_off,
 					       OCFS2_FT_DIR);
@@ -434,46 +573,8 @@ main(int argc, char **argv)
 	if (!s->quiet)
 		printf("done\n");
 
-	if (!s->hb_dev) {
-		/* These routines use libocfs2 to do their work. We
-		 * don't share an ocfs2_filesys context between the
-		 * journal format and the lost+found create so that
-		 * the library can use the journal for the latter in
-		 * future revisions. */
-
-		if (!s->no_backup_super) {
-			if (!s->quiet)
-				printf("Writing backup superblock: ");
-
-			num = format_backup_super(s);
-			if (!s->quiet)
-				printf("%d block(s)\n", num);
-		}
-
-		if (!s->quiet)
-			printf("Formatting Journals: ");
-
-		format_journals(s);
-
-		if (!s->quiet)
-			printf("done\n");
-
-		if (!s->quiet)
-			printf("Formatting slot map: ");
-
-		format_slotmap(s);
-
-		if (!s->quiet)
-			printf("done\n");
-
-		if (!s->quiet)
-			printf("Writing lost+found: ");
-
-		create_lost_found_dir(s);
-
-		if (!s->quiet)
-			printf("done\n");
-	}
+	if (!s->hb_dev)
+		finish_normal_format(s);
 
 	close_device(s);
 
@@ -523,6 +624,7 @@ get_state(int argc, char **argv)
 	int ret;
 	uint64_t val;
 	uint64_t journal_size_in_bytes = 0;
+	int journal64 = 0;
 	enum ocfs2_fs_types fs_type = FS_DEFAULT;
 	int mount = -1;
 	int no_backup_super = -1;
@@ -645,7 +747,8 @@ get_state(int argc, char **argv)
 
 		case 'J':
 			parse_journal_opts(progname, optarg,
-					   &journal_size_in_bytes);
+					   &journal_size_in_bytes,
+					   &journal64);
 			break;
 
 		case 'H':
@@ -789,6 +892,7 @@ get_state(int argc, char **argv)
 	s->format_time   = time(NULL);
 
 	s->journal_size_in_bytes = journal_size_in_bytes;
+	s->journal64 = journal64;
 
 	s->hb_dev = hb_dev;
 
@@ -809,6 +913,10 @@ get_state(int argc, char **argv)
 		s->mount = MOUNT_LOCAL;
 	else
 		s->mount = MOUNT_CLUSTER;
+ 	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_INLINE_DATA)
+ 		s->inline_data = 1;
+ 	else
+ 		s->inline_data = 0;
 	if (s->feature_flags.opt_compat & OCFS2_FEATURE_COMPAT_BACKUP_SB)
 		s->no_backup_super = 0;
 	else
@@ -883,23 +991,31 @@ get_number(char *arg, uint64_t *res)
 /* derived from e2fsprogs */
 static void
 parse_journal_opts(char *progname, const char *opts,
-		   uint64_t *journal_size_in_bytes)
+		   uint64_t *journal_size_in_bytes, int *journal64)
 {
 	char *options, *token, *next, *p, *arg;
 	int ret, journal_usage = 0;
 	uint64_t val;
+	int invert;
 
 	options = strdup(opts);
 
 	for (token = options; token && *token; token = next) {
 		p = strchr(token, ',');
 		next = NULL;
+		invert = 0;
 
 		if (p) {
 			*p = '\0';
 			next = p + 1;
 		}
 
+		arg = strstr(token, "no");
+		if (arg == token) {
+			invert = 1;
+			token += strlen("no");
+		}
+
 		arg = strchr(token, '=');
 
 		if (arg) {
@@ -908,7 +1024,7 @@ parse_journal_opts(char *progname, const char *opts,
 		}
 
 		if (strcmp(token, "size") == 0) {
-			if (!arg) {
+			if (!arg || invert) {
 				journal_usage++;
 				continue;
 			}
@@ -925,6 +1041,18 @@ parse_journal_opts(char *progname, const char *opts,
 			}
 
 			*journal_size_in_bytes = val;
+		} else if (strcmp(token, "block32") == 0) {
+			if (arg) {
+				journal_usage++;
+				continue;
+			}
+			*journal64 = invert;
+		} else if (strcmp(token, "block64") == 0) {
+			if (arg) {
+				journal_usage++;
+				continue;
+			}
+			*journal64 = !invert;
 		} else
 			journal_usage++;
 	}
@@ -933,7 +1061,9 @@ parse_journal_opts(char *progname, const char *opts,
 		com_err(progname, 0,
 			"Bad journal options specified. Valid journal "
 			"options are:\n"
-			"\tsize=<journal size>\n");
+			"\tsize=<journal size>\n"
+			"\t[no]block32\n"
+			"\t[no]block64\n");
 		exit(1);
 	}
 
@@ -1706,6 +1836,8 @@ add_entry_to_directory(State *s, DirData *dir, char *name, uint64_t byte_off,
 	de = (struct ocfs2_dir_entry *)p;
 	de->inode = 0;
 	de->rec_len = s->blocksize;
+	if (!s->inline_data || !dir->record->dir_data)
+		mkfs_init_dir_trailer(s, dir, p);
 
 got_it:
 	de->name_len = strlen(name);
@@ -1756,16 +1888,12 @@ sys_blocks_needed(uint32_t num_slots)
 static uint32_t
 system_dir_blocks_needed(State *s)
 {
-	int bytes_needed = 0;
 	int each = OCFS2_DIR_REC_LEN(SYSTEM_FILE_NAME_MAX);
 	int entries_per_block = s->blocksize / each;
 
-	bytes_needed = ((sys_blocks_needed(s->initial_slots) +
-			 entries_per_block - 1) / entries_per_block) << s->blocksize_bits;
-
-	return (bytes_needed + s->cluster_size - 1) >> s->cluster_size_bits;
+	return (sys_blocks_needed(s->initial_slots) +
+			entries_per_block - 1) / entries_per_block;
 }
-
 #if 0
 /* This breaks stuff that depends on volume_size_in_clusters and
  * volume_size_in_blocks, and I'm not even sure it's necessary. If
@@ -1799,13 +1927,21 @@ check_32bit_blocks(State *s)
 {
 	uint64_t max = UINT32_MAX;
        
+	if (s->journal64)
+		return;
+
 	if (s->volume_size_in_blocks <= max)
 		return;
 
 	fprintf(stderr, "ERROR: jbd can only store block numbers in 32 bits. "
-		"%s can hold %"PRIu64" blocks which overflows this limit. "
-		"Consider increasing the block size or decreasing the device "
-		"size.\n", s->device_name, s->volume_size_in_blocks);
+		"%s can hold %"PRIu64" blocks which overflows this limit. If "
+		"you have a new enough Ocfs2 with JBD2 support, you can try "
+		"formatting with the \"-Jblock64\" option to turn on support "
+		"for this size block device.\n"
+		"Otherwise, consider increasing the block size or "
+		"decreasing the device size.\n",
+		s->device_name, s->volume_size_in_blocks);
+
 	exit(1);
 }
 
@@ -1850,7 +1986,7 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 	if (s->hb_dev) {
 		s->feature_flags.opt_incompat =
 				 	OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV;
-		s->feature_flags.opt_compat = 0;
+		s->feature_flags.opt_compat = OCFS2_FEATURE_COMPAT_JBD2_SB;
 		s->feature_flags.opt_ro_compat = 0;
 	}
 
@@ -1874,7 +2010,13 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 	 * "s->no_backup_super" according to the features in get_state,
 	 * so it is safe to clear the flag here.
 	 */
-	s->feature_flags.opt_compat &= !OCFS2_FEATURE_COMPAT_BACKUP_SB;
+	s->feature_flags.opt_compat &= ~OCFS2_FEATURE_COMPAT_BACKUP_SB;
+
+	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_XATTR) {
+		di->id2.i_super.s_xattr_inline_size =
+						OCFS2_MIN_XATTR_INLINE_SIZE;
+		di->id2.i_super.s_uuid_hash = ocfs2_xattr_uuid_hash(s->uuid);
+	}
 
 	di->id2.i_super.s_feature_incompat = s->feature_flags.opt_incompat;
 	di->id2.i_super.s_feature_compat = s->feature_flags.opt_compat;
@@ -1883,7 +2025,8 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 	strcpy((char *)di->id2.i_super.s_label, s->vol_label);
 	memcpy(di->id2.i_super.s_uuid, s->uuid, 16);
 
-	ocfs2_swap_inode_from_cpu(di);
+	ocfs2_swap_inode_from_cpu(di, s->blocksize);
+	mkfs_compute_meta_ecc(s, di, &di->i_check);
 	do_pwrite(s, di, s->blocksize, super_off);
 	free(di);
 }
@@ -2024,10 +2167,32 @@ format_file(State *s, SystemFileDiskRecord *rec)
 		ocfs2_set_rec_clusters(0, &di->id2.i_list.l_recs[0], clusters);
 		di->id2.i_list.l_recs[0].e_blkno =
 			rec->extent_off >> s->blocksize_bits;
+	} else if (S_ISDIR(di->i_mode) && s->inline_data && rec->dir_data) {
+		DirData *dir = rec->dir_data;
+		struct ocfs2_dir_entry *de =
+			(struct ocfs2_dir_entry *)(dir->buf + dir->last_off);
+		int dir_len = dir->last_off + OCFS2_DIR_REC_LEN(de->name_len);
+
+		if (dir_len > ocfs2_max_inline_data(s->blocksize)) {
+			com_err(s->progname, 0,
+				"Inline a dir which shouldn't be inline.\n");
+			clear_both_ends(s);
+			exit(1);
+		}
+		de->rec_len -= s->blocksize -
+			       ocfs2_max_inline_data(s->blocksize);
+		memset(&di->id2, 0,
+		       s->blocksize - offsetof(struct ocfs2_dinode, id2));
+
+		di->id2.i_data.id_count = ocfs2_max_inline_data(s->blocksize);
+		memcpy(di->id2.i_data.id_data, dir->buf, dir_len);
+		di->i_dyn_features |= OCFS2_INLINE_DATA_FL;
+		di->i_size = ocfs2_max_inline_data(s->blocksize);
 	}
 
 write_out:
-	ocfs2_swap_inode_from_cpu(di);
+	ocfs2_swap_inode_from_cpu(di, s->blocksize);
+	mkfs_compute_meta_ecc(s, di, &di->i_check);
 	do_pwrite(s, di, s->blocksize, rec->fe_off);
 	free(di);
 }
@@ -2053,7 +2218,7 @@ write_bitmap_data(State *s, AllocBitmap *bitmap)
 {
 	int i;
 	uint64_t parent_blkno;
-	struct ocfs2_group_desc *gd;
+	struct ocfs2_group_desc *gd, *gd_buf;
 	char *buf = NULL;
 
 	buf = do_malloc(s, s->cluster_size);
@@ -2070,7 +2235,9 @@ write_bitmap_data(State *s, AllocBitmap *bitmap)
 		 * blkno until now. */
 		gd->bg_parent_dinode = parent_blkno;
 		memcpy(buf, gd, s->blocksize);
-		ocfs2_swap_group_desc((struct ocfs2_group_desc *)buf);
+		gd_buf = (struct ocfs2_group_desc *)buf;
+		ocfs2_swap_group_desc(gd_buf);
+		mkfs_compute_meta_ecc(s, buf, &gd_buf->bg_check);
 		do_pwrite(s, buf, s->cluster_size,
 			  gd->bg_blkno << s->blocksize_bits);
 	}
@@ -2082,20 +2249,63 @@ write_group_data(State *s, AllocGroup *group)
 {
 	uint64_t blkno = group->gd->bg_blkno;
 	ocfs2_swap_group_desc(group->gd);
+	mkfs_compute_meta_ecc(s, group->gd, &group->gd->bg_check);
 	do_pwrite(s, group->gd, s->blocksize, blkno << s->blocksize_bits);
 	ocfs2_swap_group_desc(group->gd);
 }
 
+static void mkfs_swap_dir(State *s, DirData *dir,
+			  errcode_t (*swap_entry_func)(void *buf,
+						       uint64_t bytes))
+{
+	char *p = dir->buf;
+	unsigned int offset = 0;
+	unsigned int end = s->blocksize;
+	char super_buf[OCFS2_MAX_BLOCKSIZE];
+	ocfs2_filesys fake_fs;
+	struct ocfs2_dir_block_trailer *trailer;
+
+	if (!dir->record->extent_len)
+		return;
+
+	fill_fake_fs(s, &fake_fs, super_buf);
+	if (!s->inline_data || !dir->record->dir_data)
+		if (ocfs2_supports_dir_trailer(&fake_fs))
+			end = ocfs2_dir_trailer_blk_off(&fake_fs);
+
+	while (offset < dir->record->file_size) {
+		trailer = ocfs2_dir_trailer_from_block(&fake_fs, p);
+		swap_entry_func(p, end);
+		if (end != s->blocksize)
+			ocfs2_swap_dir_trailer(trailer);
+		/* Remember, this does nothing if the feature isn't set */
+		ocfs2_compute_meta_ecc(&fake_fs, p, &trailer->db_check);
+		offset += s->blocksize;
+		p += s->blocksize;
+	}
+}
+
+static void mkfs_swap_dir_from_cpu(State *s, DirData *dir)
+{
+	mkfs_swap_dir(s, dir, ocfs2_swap_dir_entries_from_cpu);
+}
+
+static void mkfs_swap_dir_to_cpu(State *s, DirData *dir)
+{
+	mkfs_swap_dir(s, dir, ocfs2_swap_dir_entries_to_cpu);
+}
+
 static void
 write_directory_data(State *s, DirData *dir)
 {
+	if (!dir->record->extent_len)
+		return;
+
 	if (dir->buf)
-		ocfs2_swap_dir_entries_from_cpu(dir->buf,
-						dir->record->file_size);
+		mkfs_swap_dir_from_cpu(s, dir);
 	write_metadata(s, dir->record, dir->buf);
 	if (dir->buf)
-		ocfs2_swap_dir_entries_to_cpu(dir->buf,
-					      dir->record->file_size);
+		mkfs_swap_dir_to_cpu(s, dir);
 }
 
 static void
@@ -2149,17 +2359,16 @@ close_device(State *s)
 static int
 initial_slots_for_volume(uint64_t size)
 {
-	int i, shift = ONE_GB_SHIFT;
-	int defaults[4] = { 2, 4, 8, 16 };
+	size >>= ONE_GB_SHIFT;
 
-	for (i = 0, shift = ONE_GB_SHIFT; i < 4; i++, shift += 3) {
-		size >>= shift;
-
-		if (!size)
-			break;
-	}
-
-	return (i < 4) ? defaults[i] : 32;
+	if (size < 2)
+		return 2;
+	else if (size < 10)
+		return 4;
+	else if (size < 1024)
+		return 8;
+	else
+		return 16;
 }
 
 static void
@@ -2280,25 +2489,18 @@ clear_both_ends(State *s)
 	return ;
 }
 
-static void create_lost_found_dir(State *s)
+static void create_lost_found_dir(State *s, ocfs2_filesys *fs)
 {
 	errcode_t ret;
-	ocfs2_filesys *fs = NULL;
 	uint64_t lost_found_blkno;
 
-	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
-	if (ret) {
-		com_err(s->progname, ret, "while opening new file system");
-		goto bail;
-	}
-
 	ret = ocfs2_new_inode(fs, &lost_found_blkno, S_IFDIR | 0755);
 	if (ret) {
 		com_err(s->progname, ret, "while creating lost+found");
 		goto bail;
 	}
 
-	ret = ocfs2_expand_dir(fs, lost_found_blkno, fs->fs_root_blkno);
+	ret = ocfs2_init_dir(fs, lost_found_blkno, fs->fs_root_blkno);
 	if (ret) {
 		com_err(s->progname, ret, "while adding lost+found dir data");
 		goto bail;
@@ -2312,7 +2514,6 @@ static void create_lost_found_dir(State *s)
 		goto bail;
 	}
 
-	ocfs2_close(fs);
 	return ;
 
 bail:
@@ -2320,20 +2521,17 @@ bail:
 	exit(1);
 }
 
-static void format_journals(State *s)
+static void format_journals(State *s, ocfs2_filesys *fs)
 {
 	errcode_t ret;
 	int i;
 	uint32_t journal_size_in_clusters;
 	uint64_t blkno;
-	ocfs2_filesys *fs = NULL;
 	char jrnl_file[40];
-
-	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
-	if (ret) {
-		com_err(s->progname, ret, "while opening new file system");
-		goto error;
-	}
+	ocfs2_fs_options features = {
+		.opt_incompat =
+			s->journal64 ? JBD2_FEATURE_INCOMPAT_64BIT : 0,
+	};
 
 	journal_size_in_clusters = s->journal_size_in_bytes >>
 		OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits;
@@ -2350,7 +2548,8 @@ static void format_journals(State *s)
 			goto error;
 		}
 
-		ret = ocfs2_make_journal(fs, blkno, journal_size_in_clusters);
+		ret = ocfs2_make_journal(fs, blkno, journal_size_in_clusters,
+					 &features);
 		if (ret) {
 			com_err(s->progname, ret,
 				"while formatting journal \"%.*s\"",
@@ -2359,7 +2558,6 @@ static void format_journals(State *s)
 		}
 	}
 
-	ocfs2_close(fs);
 	return;
 
 error:
@@ -2367,45 +2565,24 @@ error:
 	exit(1);
 }
 
-static void format_slotmap(State *s)
+static void format_slotmap(State *s, ocfs2_filesys *fs)
 {
 	errcode_t ret;
-	ocfs2_filesys *fs;
-
-	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
-	if (ret) {
-		com_err(s->progname, ret, "while opening new file system");
-		goto error;
-	}
 
 	ret = ocfs2_format_slot_map(fs);
 	if (ret) {
-	    com_err(s->progname, ret, "while formatting the slot map");
-	    goto error;
+		com_err(s->progname, ret, "while formatting the slot map");
+		clear_both_ends(s);
+		exit(1);
 	}
-
-	ocfs2_close(fs);
-	return;
-
-error:
-	clear_both_ends(s);
-	exit(1);
 }
 
-static int format_backup_super(State *s)
+static int format_backup_super(State *s, ocfs2_filesys *fs)
 {
 	errcode_t ret;
-	ocfs2_filesys *fs = NULL;
 	size_t len;
 	uint64_t blocks[OCFS2_MAX_BACKUP_SUPERBLOCKS];
 
-	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
-	if (ret) {
-		com_err(s->progname, ret,
-			"while opening file system for backup superblock.");
-		goto error;
-	}
-
 	len = ocfs2_get_backup_super_offsets(fs, blocks, ARRAY_SIZE(blocks));
 
 	ret = ocfs2_set_backup_super_list(fs, blocks, len);
@@ -2424,10 +2601,16 @@ static int format_backup_super(State *s)
 		goto error;
 	}
 
-	ocfs2_close(fs);
 	return len;
 
 error:
 	clear_both_ends(s);
 	exit(1);
 }
+
+static void mkfs_compute_meta_ecc(State *s, void *data,
+				  struct ocfs2_block_check *bc)
+{
+	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_META_ECC)
+		ocfs2_block_check_compute(data, s->blocksize, bc);
+}
